//----------------------------------------------------------------------------------------
// boundary_data.cpp: Boundary counditions for different types of boundary edges/faces
//----------------------------------------------------------------------------------------

#include <iostream>
#include "general.H"
#include "mesh_2d_sw.H"
#include "mesh_3d_sw.H"
#include "mesh_container.H"
#include "pde.H"
#ifdef TEST_KEY
#include "test.H"
#endif

extern PROBLEM_DESCRIPTION problem_info;   // Storage for problem parameters
extern MESH mesh;                          // Mesh object
#if TURBULENCE_KEY==3  // two equation turbulence models
extern GLS_Data gls_data;                  // Data for two equation turbulence models
#endif


//--------------------------------------------------------------------------------------
//                           Edges
//--------------------------------------------------------------------------------------


// Update often changed data at quadrature points (unknowns)
// Parameters: t - time
void EDGE_LAND::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  var[XI]=quad_local().Var[0][VAR_2D_DT_INDEX_INV[XI]];
}

// Update often changed data at quadrature points (unknowns)
// Parameters: t - time
void EDGE_SEA::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  var[XI]=TidalElevation();
}

// Update often changed data at quadrature points (unknowns)
// Parameters: t - time
void EDGE_RIVER::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  var[XI]=SurfaceElevation();
}

// Update often changed data at quadrature points (unknowns)
// Parameters: t - time
void EDGE_RADIATION::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  var[XI]=quad_local().Var[0][VAR_2D_DT_INDEX_INV[XI]];
}

#ifdef TEST_KEY
// Update dynamic quantities at the Gauss points
void EDGE_UNKNOWN::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  solution_analytic_elevation(t, coord_quad(), &var[XI]);
}
#endif



//--------------------------------------------------------------------------------------
//                          Horizontal faces
//--------------------------------------------------------------------------------------


// Compute boundary conditions for primary variables
// Parameters: t - time; param - parameter list; var - result
void FACE_HORIZ_SW::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  static const VAR* ind=VAR_3D_DT_INDEX;
  int k = IsBottom() ? 0 : 1;

  for (int i=0; i<d_var_3d_dt; i++)
    var[ind[i]]=Var_dof(k, ind[i])*basis_func_quad(k, ind[i]);
  var[W]=-(var[U]*Normal(0)+var[V]*Normal(1))/Normal(2);
}

// Compute flux (auxilliary variable) at a quadrature point
// Parameters: t - time; diff_coef - diffusion coefficients; diff - result
void FACE_HORIZ_SW::BC_Neumann_quad(real t, const real diff_coef[d_var][d_space],
                                    real diff[d_var][d_space])
{
  int k = IsBottom() ? 0 : 1;

  for (int i=0; i<problem_info.n_diff_horiz; i++)
  {
    VAR ind=problem_info.diff_horiz_ind[i];
    diff[ind][0]=diff_coef[ind][0]*(Q_dof(k, ind, 0)*basis_func_quad(k, ind));
    diff[ind][1]=diff_coef[ind][1]*(Q_dof(k, ind, 1)*basis_func_quad(k, ind));
  }
  for (int i=0; i<problem_info.n_diff_vert; i++)
  {
    VAR ind=problem_info.diff_vert_ind[i];
    diff[ind][2]=diff_coef[ind][2]*(Q_dof(k, ind, 2)*basis_func_quad(k, ind));
  }
}

// Compute boundary conditions for primary variables
// Parameters: t - time; param - parameter list; var - result
void FACE_HORIZ_SURFACE::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  static const VAR* ind=VAR_3D_DT_INDEX;
  for (int i=0; i<d_var_3d_dt; i++)
    var[ind[i]]=Var_dof(1, ind[i])*basis_func_quad(1, ind[i]);
  var[W]=Var_dof(1, W)*basis_func_quad(1, W);
}

// Compute flux (auxilliary variable) at a quadrature point
// Parameters: t - time; diff_coef - diffusion coefficients; diff - result
void FACE_HORIZ_SURFACE::BC_Neumann_quad(real t, const real diff_coef[d_var][d_space],
                                         real diff[d_var][d_space])
{
#ifdef TEST_KEY
  real q_app[d_var][d_space]={{0.}};

  solution_analytic_grad(t, coord_quad(), q_app);
  for (int i=0; i<problem_info.n_diff_horiz; i++)
  {
    VAR ind=problem_info.diff_horiz_ind[i];
    diff[ind][0]=-diff_coef[ind][0]*q_app[ind][0];
    diff[ind][1]=-diff_coef[ind][1]*q_app[ind][1];
  }
  for (int i=0; i<problem_info.n_diff_vert; i++)
  {
    VAR ind=problem_info.diff_vert_ind[i];
    diff[ind][2]=-diff_coef[ind][2]*q_app[ind][2];
  }
#else
  for (int i=0; i<problem_info.n_diff_horiz; i++)
  {
    VAR ind=problem_info.diff_horiz_ind[i];
    diff[ind][0]=diff[ind][1]=0.;
  }
  for (int i=0; i<problem_info.n_diff_vert; i++)
  {
    VAR ind=problem_info.diff_vert_ind[i];
    diff[ind][2]=0.;
  }
#endif
}


// Compute boundary conditions for primary variables
// Parameters: t - time; param - parameter list; var - result
void FACE_HORIZ_SURFACE_STRESS::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  static const VAR* ind=VAR_3D_DT_INDEX;
  for (int i=0; i<d_var_3d_dt; i++)
    var[ind[i]]=Var_dof(1, ind[i])*basis_func_quad(1, ind[i]);
  var[W]=Var_dof(1, W)*basis_func_quad(1, W);
}

// Compute flux (auxilliary variable) at a quadrature point
// Parameters: t - time; diff_coef - diffusion coefficients; diff - result
void FACE_HORIZ_SURFACE_STRESS::BC_Neumann_quad(real t, const real diff_coef[d_var][d_space],
                                                real diff[d_var][d_space])
{
  for (int i=0; i<problem_info.n_diff_horiz; i++)
  {
    VAR ind=problem_info.diff_horiz_ind[i];
    diff[ind][0]=diff[ind][1]=0.;
  }
  for (int i=0; i<problem_info.n_diff_vert; i++)
  {
    VAR ind=problem_info.diff_vert_ind[i];
    diff[ind][2]=0.;
  }
  diff[U][2]=0.5*problem_info.ramp_value*(diff[U][2]-problem_info.surface_stress[0]);
  diff[V][2]=0.5*problem_info.ramp_value*(diff[V][2]-problem_info.surface_stress[1]);
#if TURBULENCE_KEY>1
  if (problem_info.diff_vert_switch>3)
    diff[K][2]=0.5*diff[K][2];
#if TURBULENCE_KEY==3  // two equation turbulence models
  if (problem_info.diff_vert_switch>4)
    {
      real k=Var_dof(1, K)*basis_func_quad(1, K);
      k = k>gls_data.k_min ? k : gls_data.k_min;
      real u=Var_dof(1, U)*basis_func_quad(1, U), v=Var_dof(1, V)*basis_func_quad(1, V);
      real tau_s=std::sqrt(diff[U][2]*diff[U][2]+diff[V][2]*diff[V][2]);
      tau_s = tau_s<1.e-8 ? 1.e-8 : tau_s;
      real l=problem_info.surface_rough_coef*std::exp(gls_data.kappa*std::sqrt((u*u+v*v)/tau_s));
      diff[E][2]=0.5*(diff[E][2]+diff_coef[E][2]*gls_data.Psi_BC(k, l));
    }
#endif
#endif
}


// Compute boundary conditions for primary variables
// Parameters: t - time; param - parameter list; var - result
void FACE_HORIZ_BOTTOM_NO_SLIP::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  static const VAR* ind=VAR_3D_DT_INDEX;
  for (int i=0; i<d_var_3d_dt; i++)
    var[ind[i]]=Var_dof(0, ind[i])*basis_func_quad(0, ind[i]);
  var[U]=var[V]=var[W]=0.;
}

// Compute flux (Neumann) boundary conditions at a quadrature point
// Parameters: t - time; diff_coef - diffusion coefficients; diff - result
void FACE_HORIZ_BOTTOM_NO_SLIP::BC_Neumann_quad(real t, const real diff_coef[d_var][d_space], 
                                                real diff[d_var][d_space])
{
  FACE_HORIZ_SW::BC_Neumann_quad(t, diff_coef, diff);
#if TURBULENCE_KEY>1
  if (problem_info.diff_vert_switch>3)
    diff[K][2]=0.5*diff[K][2];
#if TURBULENCE_KEY==3  // two equation turbulence models
  if (problem_info.diff_vert_switch>4)
  {
    real u=Var_dof(0, U)*basis_func_quad(0, U), v=Var_dof(0, V)*basis_func_quad(0, V), 
      U_abs=std::sqrt(u*u+v*v);
    real k=Var_dof(0, K)*basis_func_quad(0, K);
    k = k>gls_data.k_min ? k : gls_data.k_min;
    real tau_s=std::pow(diff[U][2]*diff[U][2]+diff[V][2]*diff[V][2], (real)0.25);
    tau_s = tau_s<1.e-6 ? 1.e-6 : tau_s;
    real l=problem_info.bottom_rough_coef*std::exp(gls_data.kappa*U_abs/tau_s);
    diff[E][2]=0.5*(diff[E][2]-diff_coef[E][2]*gls_data.Psi_BC(k, l));
  }
#endif
#endif
}


// Compute boundary conditions for primary variables
// Parameters: t - time; param - parameter list; var - result
void FACE_HORIZ_BOTTOM_NO_DRAG::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  static const VAR* ind=VAR_3D_DT_INDEX;
  for (int i=0; i<d_var_3d_dt; i++)
    var[ind[i]]=Var_dof(0, ind[i])*basis_func_quad(0, ind[i]);
  var[W]=-(var[U]*Normal(0)+var[V]*Normal(1))/Normal(2);
}

// Compute flux (Neumann) boundary conditions at a quadrature point
// Parameters: t - time; diff_coef - diffusion coefficients; diff - result
void FACE_HORIZ_BOTTOM_NO_DRAG::BC_Neumann_quad(real t, const real diff_coef[d_var][d_space],
                                                real diff[d_var][d_space])
{
  for (int i=0; i<problem_info.n_diff_horiz; i++)
  {
    VAR ind=problem_info.diff_horiz_ind[i];
    diff[ind][0]=diff[ind][1]=0.;
  }
  for (int i=0; i<problem_info.n_diff_vert; i++)
  {
    VAR ind=problem_info.diff_vert_ind[i];
    diff[ind][2]=0.;
  }
#if TURBULENCE_KEY>1
  if (problem_info.diff_vert_switch>3)
    diff[K][2]=0.5*diff[K][2];
#if TURBULENCE_KEY==3  // two equation turbulence models
  if (problem_info.diff_vert_switch>4)
  {
    real u=Var_dof(0, U)*basis_func_quad(0, U), v=Var_dof(0, V)*basis_func_quad(0, V), 
      U_abs=std::sqrt(u*u+v*v);
    real k=Var_dof(0, K)*basis_func_quad(0, K);
    k = k>gls_data.k_min ? k : gls_data.k_min;
    real tau_s=std::pow(diff[U][2]*diff[U][2]+diff[V][2]*diff[V][2], (real)0.25);
    tau_s = tau_s<1.e-6 ? 1.e-6 : tau_s;
    real l=problem_info.bottom_rough_coef*std::exp(gls_data.kappa*U_abs/tau_s);
    diff[E][2]=0.5*(diff[E][2]-diff_coef[E][2]*gls_data.Psi_BC(k, l));
  }
#endif
#endif
}


// Compute boundary conditions for primary variables
// Parameters: t - time; param - parameter list; var - result
void FACE_HORIZ_BOTTOM_QUAD_DRAG::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  static const VAR* ind=VAR_3D_DT_INDEX;

  for (int i=0; i<d_var_3d_dt; i++)
    var[ind[i]]=Var_dof(0, ind[i])*basis_func_quad(0, ind[i]);
  var[W]=-(var[U]*Normal(0)+var[V]*Normal(1))/Normal(2);
}

// Compute flux (Neumann) boundary conditions at a quadrature point
// Parameters: t - time; diff_coef - diffusion coefficients; diff - result
void FACE_HORIZ_BOTTOM_QUAD_DRAG::BC_Neumann_quad(real t, const real diff_coef[d_var][d_space],
                                                  real diff[d_var][d_space])
{
  FACE_HORIZ_SW::BC_Neumann_quad(t, diff_coef, diff);
  real u=Var_dof(0, U)*basis_func_quad(0, U), v=Var_dof(0, V)*basis_func_quad(0, V), 
    U_abs=std::sqrt(u*u+v*v);
  diff[U][2]=0.5*(diff[U][2]-problem_info.bottom_fric_coef*U_abs*u);
  diff[V][2]=0.5*(diff[V][2]-problem_info.bottom_fric_coef*U_abs*v);
#if TURBULENCE_KEY>1
  if (problem_info.diff_vert_switch>3)
    diff[K][2]=0.5*diff[K][2];
#if TURBULENCE_KEY==3  // two equation turbulence models
  if (problem_info.diff_vert_switch>4)
    {
      real k=Var_dof(0, K)*basis_func_quad(0, K);
      k = k>gls_data.k_min ? k : gls_data.k_min;
      real tau_s=std::pow(diff[U][2]*diff[U][2]+diff[V][2]*diff[V][2], (real)0.25);
      tau_s = tau_s<1.e-6 ? 1.e-6 : tau_s;
      real l=problem_info.bottom_rough_coef*std::exp(gls_data.kappa*U_abs/tau_s);
      diff[E][2]=0.5*(diff[E][2]-diff_coef[E][2]*gls_data.Psi_BC(k, l));
    }
#endif
#endif
}


// Compute boundary conditions for primary variables
// Parameters: t - time; param - parameter list; var - result
void FACE_HORIZ_BOTTOM_LIN_DRAG::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  static const VAR* ind=VAR_3D_DT_INDEX;
  for (int i=0; i<d_var_3d_dt; i++)
    var[ind[i]]=Var_dof(0, ind[i])*basis_func_quad(0, ind[i]);
  var[W]=-(var[U]*Normal(0)+var[V]*Normal(1))/Normal(2);
}

// Compute flux (Neumann) boundary conditions at a quadrature point
// Parameters: t - time; diff_coef - diffusion coefficients; diff - result
void FACE_HORIZ_BOTTOM_LIN_DRAG::BC_Neumann_quad(real t, const real diff_coef[d_var][d_space],
                                                 real diff[d_var][d_space])
{
  FACE_HORIZ_SW::BC_Neumann_quad(t, diff_coef, diff);

  real u=Var_dof(0, U)*basis_func_quad(0, U), v=Var_dof(0, V)*basis_func_quad(0, V);
  diff[U][2]=-problem_info.bottom_fric_coef*u;
  diff[V][2]=-problem_info.bottom_fric_coef*v;
#if TURBULENCE_KEY>1
  if (problem_info.diff_vert_switch>3)
    diff[K][2]=0.5*diff[K][2];
#if TURBULENCE_KEY==3  // two equation turbulence models
  if (problem_info.diff_vert_switch>4)
  {
    real U_abs=std::sqrt(u*u+v*v);
    real k=Var_dof(0, K)*basis_func_quad(0, K);
    k = k>gls_data.k_min ? k : gls_data.k_min;
    real tau_s=std::pow(diff[U][2]*diff[U][2]+diff[V][2]*diff[V][2], (real)0.25);
    tau_s = tau_s<1.e-6 ? 1.e-6 : tau_s;
    real l=problem_info.bottom_rough_coef*std::exp(gls_data.kappa*U_abs/tau_s);
    diff[E][2]=0.5*(diff[E][2]-diff_coef[E][2]*gls_data.Psi_BC(k, l));
  }
#endif
#endif
}


#ifdef TEST_KEY
// Compute boundary conditions for primary variables
// Parameters: t - time; param - parameter list; var - result
void FACE_HORIZ_BOTTOM_TEST::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  solution_analytic(t, coord_quad(), var);
}

// Compute flux (Neumann) boundary conditions at a quadrature point
// Parameters: t - time; diff_coef - diffusion coefficients; diff - result
void FACE_HORIZ_BOTTOM_TEST::BC_Neumann_quad(real t, const real diff_coef[d_var][d_space],
                                             real diff[d_var][d_space])
{
  real q_app[d_var][d_space]={{0.}};

  solution_analytic_grad(t, coord_quad(), q_app);
  for (int i=0; i<problem_info.n_diff_horiz; i++)
  {
    VAR ind=problem_info.diff_horiz_ind[i];
    diff[ind][0]=-diff_coef[ind][0]*q_app[ind][0];
    diff[ind][1]=-diff_coef[ind][1]*q_app[ind][1];
  }
  for (int i=0; i<problem_info.n_diff_vert; i++)
  {
    VAR ind=problem_info.diff_vert_ind[i];
    diff[ind][2]=-diff_coef[ind][2]*q_app[ind][2];
  }
}
#endif



//--------------------------------------------------------------------------------------
//                          Vertical faces
//--------------------------------------------------------------------------------------

// Compute flux (Neumann) boundary conditions at a quadrature point
// Parameters: t - time; diff_coef - diffusion coefficients; diff - result
void FACE_VERT_SW::BC_Neumann_quad(real t, const real diff_coef[d_var][d_space],
                                   real diff[d_var][d_space])
{
  for (int i=0; i<problem_info.n_diff_horiz; i++)
  {
    VAR ind=problem_info.diff_horiz_ind[i];
    diff[ind][0]=diff[ind][1]=0.;
  }
}

// Compute boundary conditions for primary variables
// Parameters: t - time; param - parameter list; var - result
void FACE_VERT_RIVER::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  static const VAR* ind=VAR_TRANSPORT_INDEX;
  EDGE_RIVER* container_river=(EDGE_RIVER*)container_2d;
  real vel_scale=0., vel[d_space_2d]={problem_info.ramp_value*(container_river->Velocity(0)*Normal(0)-
                                                               container_river->Velocity(1)*Normal(1)),
                                      problem_info.ramp_value*(container_river->Velocity(0)*Normal(1)+
                                                               container_river->Velocity(1)*Normal(0))};
// #ifdef DEPTH_AVERAGED_VELOCITY_KEY // Fitting vertical velocity profile if possible
//   real vel_av=UH_quad(0)*Normal(0)+UH_quad(1)*Normal(1);
//   if (std::abs(vel_av)>1.e-8*std::abs(container_river->Velocity(0)))
//     vel_scale=container_river->Velocity(0)/vel_av;
// #endif

//   if (vel_scale!=0.)
//     {
//       var[U]=(Var_dof(0, U)*basis_func_quad(0, U))*vel_scale;
//       var[V]=(Var_dof(0, V)*basis_func_quad(0, V))*vel_scale;
//     }
//   else
    {
      var[U]=vel[0];
      var[V]=vel[1];
    }
  for (int i=0; i<d_var_transport; i++)
    var[ind[i]]=Var_dof(0, ind[i])*basis_func_quad(0, ind[i]);
#ifdef BAROCLINIC_KEY
  var[T]=container_river->Temperature();
  var[S]=container_river->Salinity();
#endif
}

// Compute flux (Neumann) boundary conditions at a quadrature point
// Parameters: t - time; diff_coef - diffusion coefficients; diff - result
void FACE_VERT_RIVER::BC_Neumann_quad(real t, const real diff_coef[d_var][d_space],
                                real diff[d_var][d_space])
{
  for (int i=0; i<problem_info.n_diff_horiz; i++)
  {
    VAR ind=problem_info.diff_horiz_ind[i];
    diff[ind][0]=diff_coef[ind][0]*(Q_dof(0, ind, 0)*basis_func_quad(0, ind));
    diff[ind][1]=diff_coef[ind][1]*(Q_dof(0, ind, 1)*basis_func_quad(0, ind));
  }
}

// Compute boundary conditions for primary variables
// Parameters: t - time; param - parameter list; var - result
void FACE_VERT_LAND::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  static const VAR* ind=VAR_TRANSPORT_INDEX;
  real u=Var_dof(0, U)*basis_func_quad(0, U), v=Var_dof(0, V)*basis_func_quad(0, V); 
  long_real upwind=u*Normal(0)+v*Normal(1);

// #ifdef SUPERCRITICAL_KEY
  var[U]=u-2.*upwind*Normal(0);
  var[V]=v-2.*upwind*Normal(1);
// #else
//   var[U]=u-upwind*Normal(0);
//   var[V]=v-upwind*Normal(1);
// #endif
  for (int i=0; i<d_var_transport; i++)
    var[ind[i]]=Var_dof(0, ind[i])*basis_func_quad(0, ind[i]);
}


// Compute boundary conditions for primary variables
// Parameters: t - time; param - parameter list; var - result
void FACE_VERT_SEA::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  static const VAR* ind=VAR_3D_DT_INDEX;

  for (int i=0; i<d_var_3d_dt; i++)
    var[ind[i]]=Var_dof(0, ind[i])*basis_func_quad(0, ind[i]);
#ifdef BAROCLINIC_KEY
  var[T]=((EDGE_SEA*)container_2d)->Temperature();
  var[S]=((EDGE_SEA*)container_2d)->Salinity();
#endif
}


// Compute boundary conditions for primary variables
// Parameters: t - time; param - parameter list; var - result
void FACE_VERT_RADIATION::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  static const VAR* ind=VAR_3D_DT_INDEX;

  for (int i=0; i<d_var_3d_dt; i++)
    var[ind[i]]=Var_dof(0, ind[i])*basis_func_quad(0, ind[i]);
}

#ifdef TEST_KEY
// Compute boundary conditions for primary variables
// Parameters: t - time; param - parameter list; var - result
void FACE_VERT_UNKNOWN::BC_Dirichlet_quad(real t, const real param[], real var[d_var])
{
  solution_analytic(t, coord_quad(), var);
}

// Compute flux (Neumann) boundary conditions at a quadrature point
// Parameters: t - time; diff_coef - diffusion coefficients; diff - result
void FACE_VERT_UNKNOWN::BC_Neumann_quad(real t, const real diff_coef[d_var][d_space],
                                  real diff[d_var][d_space])
{
  real q_app[d_var][d_space]={{0.}};

  solution_analytic_grad(t, coord_quad(), q_app);
  for (int i=0; i<problem_info.n_diff_horiz; i++)
  {
    VAR ind=problem_info.diff_horiz_ind[i];
    diff[ind][0]=-diff_coef[ind][0]*q_app[ind][0];
    diff[ind][1]=-diff_coef[ind][1]*q_app[ind][1];
  }
}
#endif

