//----------------------------------------------------------------------------------------
// dof.H: Contains declarations for objects containing degrees of freedom for unknowns
//----------------------------------------------------------------------------------------

#ifndef DOF_H
#define DOF_H

#include <fstream>
#include <vector>
#include "general.H"

class DOF_iterator;


//----------------------------------------------------------------------------------------
// class DOF: Defines storage for global(all elements) degrees of freedom for one unknown
//----------------------------------------------------------------------------------------

class DOF
{
private:
  unsigned char dim;          // Number of local(one element) degrees of freedom
  std::vector<real> storage;  // Storage array
  DOF_iterator* iter;         // Access iterator to degrees of freedom corresponding to one element

public:
  // Default constructor
  DOF();

  // Destructor
  ~DOF();

  // Return value: Number of local (one element) degrees of freedom
  unsigned char GetDim() const {return dim;}

  // Return value: Number of global (all elements) degrees of freedom
  std::size_t GetSize() const {return storage.size();}

  // Initialize global storage for degrees of freedom
  // Parameters: n_el - number of elements, n_dof number of local degrees of freedom
  void InitializeGlobal(int n_el, int n_dof);

  // Access storage for degrees of freedom for one element
  real* GetLocal(int el)
  {
    assert_debug(el>=0 && (el+1)*dim<=storage.size());
    return &storage[el*dim];
  }

  // Return value: Access iterator
  DOF_iterator* GetDofIterator() {return iter;}

  // Initialize storage for degrees of freedom for one element
  void InitializeLocal(int el, int n_dof) // Local storage allocation
  {
    assert_debug(el>0 && n_dof>0 && n_dof<256 && dim==n_dof);
    std::size_t new_size=(el+1)*dim;
    if (storage.size()<new_size)
      storage.resize(new_size);
  }

  // Access a single value in global array of degrees of freedom: rvalue
  // Parameters: i - Index in global storage for degrees of freedom
  // Return value: value stored in i-th cell in global storage for degrees of freedom
  real operator()(std::size_t i) const
  {
    assert_debug(i<storage.size());
    return storage[i];
  }

  // Access a single value in global array of degrees of freedom: lvalue
  // Parameters: i - Index in global storage for degrees of freedom
  // Return value: value stored in i-th cell in global storage for degrees of freedom
  real& operator()(std::size_t i)
  {
    assert_debug(i<storage.size());
    return storage[i];
  }

  // Assign constant to global degrees of freedom 
  // Parameters: d - constant value assigned to all degrees of freedom
  // Return value: d
  real operator=(real d)
  {
    std::size_t array_length=storage.size();
    for (std::size_t i=0; i<array_length; storage[i]=d, i++);
    return d;
  }

  // Global assignment operator
  // Parameters: R - DOF object to be assigned to the current object
  // Return value: R
  DOF& operator=(const DOF& R)
  {
    std::size_t array_length=storage.size();
    assert_debug(array_length==R.GetSize());
    for (std::size_t i=0; i<array_length; storage[i]=R(i), i++);
    return *this;
  }

  // Global *= operator
  // Parameters: d - value to be multiplied with
  // Return value: *this
  DOF& operator*=(long_real d)
  {
    std::size_t array_length=storage.size();
    for (std::size_t i=0; i<array_length; storage[i]*=d, i++);
    return *this;
  }

  // Global +=d*DOF operator
  // Parameters: d - value to be multiplied with R, R - another DOF object
  void Add(long_real d, const DOF& R) // Global: *this+=d*R
  {
    std::size_t array_length=storage.size();
    assert_debug(array_length==R.GetSize());
    for (std::size_t i=0; i<array_length; storage[i]+=d*R(i), i++);
  }

  // Hot file output
  friend std::ofstream& operator<<(std::ofstream& out, const DOF& dof);

  // Hot file input
  friend std::ifstream& operator>>(std::ifstream& in, DOF& dof);
};



//----------------------------------------------------------------------------------------
// class DOF_iterator: Enables access to local(one element) degrees of freedom for one unknown
//----------------------------------------------------------------------------------------

class DOF_iterator
{
private:
  unsigned char dim;  // Number of local degrees of freedom
  real *data;         // Pointer to local degrees of freedom
  DOF *dof;           // Pointer to corresponding DOF object

public:
  // Default constructor
  DOF_iterator(DOF* dof_new) {dim=dof_new->GetDim(); data=0; dof=dof_new;}

  // Constructor for a temporary local storage object
  DOF_iterator(unsigned char dim_, real *data_) {dim=dim_; data=data_; dof=0;}

  // Destructor
  ~DOF_iterator() {}

  // Get pointer to the corresponding DOF object
  DOF* GetDof() {return dof;}

  // Return value: number of degrees of freedom for one element: rvalue
  unsigned char GetDim() const {return dim;}

  // Return value: number of degrees of freedom for one element: lvalue
  unsigned char& SetDim() {return dim;}

  // Assign to an element
  // Parameters: el - element index
  void SetElement(int el)
  {
    assert_debug(dof!=0);
    data=dof->GetLocal(el);
  }

  // Local assignment operator
  // Parameters: dof - DOF_iterator object to be assigned to the current object
  // Return value: *this
  const DOF_iterator& operator=(const DOF_iterator& dof_new) const
  {
    assert_debug(data!=0);
    for (int i=0; i<dim; data[i]=dof_new(i), i++);
    return *this;
  }

  // Local assignment operator
  // Parameters: r - real[] array with values to be assigned to the local degrees of freedom
  // Return value: *this
  const DOF_iterator& operator=(const real *r) const
  {
    assert_debug(data!=0);
    for (int i=0; i<dim; data[i]=r[i], i++);
    return *this;
  }

  // Local assignment operator
  // Parameters: d - constant to be assigned to the local degrees of freedom
  // Return value: *this
  const DOF_iterator& operator=(real d) const
  {
    assert_debug(data!=0);
    for (int i=0; i<dim; data[i]=d, i++);
    return *this;
  }

  // Access a single value in local array of degrees of freedom: rvalue
  // Parameters: i - local index
  // Return value: value stored in i-th cell in local storage for degrees of freedom
  real operator()(int i) const
  {
    assert_debug(data!=0);
    return data[i];
  }

  // Access a single value in local array of degrees of freedom: lvalue
  // Parameters: i - local index
  // Return value: value stored in i-th cell in local storage for degrees of freedom
  real& operator()(int i)
  {
    assert_debug(data!=0);
    return data[i];
  }

  // Access a local array of degrees of freedom: lvalue
  // Return value: pointer to local storage for degrees of freedom
  real* operator&() const
  {
    assert_debug(data!=0);
    return data;
  }

  // Local dot-product operator
  // Parameters: r - array to be multiplied with
  // Return value: product
  long_real operator*(const real *r) const
  {
    long_real result=0.;
    assert_debug(data!=0);
    for (int i=0; i<dim; result+=data[i]*r[i], i++);
    return result;
  }

  // Local *= operator
  // Parameters: d - value to be multiplied with
  // Return value: *this
  DOF_iterator& operator*=(long_real d)
  {
    assert_debug(data!=0);
    for (int i=0; i<dim; data[i]*=d, i++);
    return *this;
  }

  // Local += operator
  // Parameters: dof_new - DOF_iterator object to be added to the current object
  // Return value: *this
  DOF_iterator& operator+=(const DOF_iterator& dof_new)
  {
    assert_debug(data!=0);
    for (int i=0; i<dim; data[i]+=dof_new(i), i++);
    return *this;
  }

  // Local +=d*DOF_iterator operator
  // Parameters: d - value to be multiplied with r, r - array of local degrees of freedom
  void Add(long_real d, const real* r) const
  {
    assert_debug(data!=0);
    for (int i=0; i<dim; data[i]+=d*r[i], i++);
  }
};

#endif
