//----------------------------------------------------------------------------------------
// general.H: Contains declarations used throughout the program
//----------------------------------------------------------------------------------------

#ifndef GENERAL_H
#define GENERAL_H

#include <cassert>

//--------------------------------------------------------------------------------------------------
// Compilation keys
//--------------------------------------------------------------------------------------------------
// #define DEBUG_KEY            // Key turning on debugging checks
// #define TEST_KEY             // Key needed for test problems with an analytical solution
// #define SUPERCRITICAL_KEY    // Key needed for supercritical flow problems
#define BAROCLINIC_KEY       // Key turning on the baroclinic model
// #define ADAPTIVITY_H_2D_KEY  // Key turning on adaptivity with the 2D shock-detector
// #define ADAPTIVITY_H_3D_KEY  // Key turning on adaptivity with the 3D shock-detector
// #define ROUNDOFF_KEY         // Key turning on roundoff reduction features
// #define MASS_CONSERVATION_KEY   // Conservation of mass computation key
#define DEPTH_AVERAGED_VELOCITY_KEY  // Enable computation of depth averaged horizontal velocity key
#define TURBULENCE_KEY    3  // Turbulence closure model choice key:
                             // 1 - algebraic, 2 - one equation, 3 - two equations


//--------------------------------------------------------------------------------------------------
// Floating type definitions
//--------------------------------------------------------------------------------------------------
typedef float real;         // Short floating point type definition
typedef float long_real;    // Long floating point type definition


//--------------------------------------------------------------------------------------------------
// Global constants
//--------------------------------------------------------------------------------------------------
const int BAD_INDEX=-1;      // Default index pointing nowhere
const int d_space=3;         // Space dimension in 3D
const int d_space_2d=2;      // Space dimension in 2D
const int max_order=2;       // Maximal order of approximation (quadratic)
const int max_dim=10;        // Maximal number of degrees of freedom in 3D
const int max_quad=27;       // Maximal number of quadrature points
const int vert_h=3;          // Number of vertices in a triangle
const int d_rk=3;            // Number of different Runge-Kutta schemes implemented
const int max_stages=3;      // Maximal number of stages in a Runge-Kutta method
const int file_name=120;     // Admissible length of a file name


//--------------------------------------------------------------------------------------------------
// Redefined assert function to be called only if DEBUG_KEY is defined
//--------------------------------------------------------------------------------------------------
#ifdef DEBUG_KEY
#define assert_debug(condition) assert(condition)
#else
#define assert_debug(condition)
#endif


//--------------------------------------------------------------------------------------------------
// Basis function and basis gradient type definitions
//--------------------------------------------------------------------------------------------------
typedef real func_type[max_dim], grad_type[max_dim][d_space];



//--------------------------------------------------------------------------------------------------
// Global indices for all unknowns
//--------------------------------------------------------------------------------------------------
enum VAR {XI, U, V, W

#ifdef BAROCLINIC_KEY  // Baroclinic case unknowns
          , T, S, RHO, RHO_FRS
#endif

#ifdef MASS_CONSERVATION_KEY  // global mass conservation unknown
, TRACER
#endif

#ifdef DEPTH_AVERAGED_VELOCITY_KEY  // algebraic turbulence model unknowns
          , UH, VH
#endif

#if TURBULENCE_KEY>1   // "k" turbulence model unknown
#ifdef BAROCLINIC_KEY  // Baroclinic case unknowns: density gradient
          , RHO_Z
#endif
          , K
#if TURBULENCE_KEY==3  // "k-epsilon" and Mellor-Yamada turbulence model unknown
          , E
#endif
#endif
          , d_var};



//--------------------------------------------------------------------------------------------------
// Automatic generation of an inverse to an index of unknowns (can be also done manually)
//--------------------------------------------------------------------------------------------------
class Inverse
{
private:
  int index_inv[d_var];

public:
  Inverse(const VAR index[], int dim)
  {
    int i, j;

    for (i=0; i<d_var; i++)
    {
      for (j=0; j<dim; j++)
        if (index[j]==i)
        {
          index_inv[i]=j;
          break;
        }
      if (j==dim)
        index_inv[i]=BAD_INDEX;
    }
  }

  int operator[](int i) const
  {
    assert_debug(i>=0 && i<d_var);
    return index_inv[i];
  }

  int operator[](VAR i) const
  {
    assert_debug(i>=0 && i<d_var);
    return index_inv[i];
  }
};



//--------------------------------------------------------------------------------------------------
// Names for all unknowns
//--------------------------------------------------------------------------------------------------
const char* const VAR_NAME[d_var]={"Xi", "U", "V", "W"
#ifdef BAROCLINIC_KEY  // Baroclinic case unknowns
                                   , "T", "S", "Rho", "Rho_forcing"
#endif

#ifdef MASS_CONSERVATION_KEY  // global mass conservation unknown
, "Tracer"
#endif

#ifdef DEPTH_AVERAGED_VELOCITY_KEY  // algebraic turbulence model unknowns
                                   , "UH", "VH"
#endif

#if TURBULENCE_KEY>1  // "k" turbulence model unknown
#ifdef BAROCLINIC_KEY  // Baroclinic case unknowns: density gradient
                                   , "Rho_z"
#endif
                                   , "K"
#if TURBULENCE_KEY==3  // "k-epsilon" and Mellor-Yamada turbulence model unknown
                                   , "E"
#endif
#endif
};


//--------------------------------------------------------------------------------------------------
// 2D unknowns
//--------------------------------------------------------------------------------------------------
const VAR VAR_2D_INDEX[]={XI

#ifdef DEPTH_AVERAGED_VELOCITY_KEY  // algebraic turbulence model unknowns
                          , UH, VH
#endif
};

const int d_var_2d = sizeof VAR_2D_INDEX / sizeof *VAR_2D_INDEX;

const Inverse VAR_2D_INDEX_INV(VAR_2D_INDEX, d_var_2d);



//--------------------------------------------------------------------------------------------------
// Time dependent 2D unknowns
//--------------------------------------------------------------------------------------------------
const VAR VAR_2D_DT_INDEX[]={XI};

const int d_var_2d_dt = sizeof VAR_2D_DT_INDEX / sizeof *VAR_2D_DT_INDEX;

const Inverse VAR_2D_DT_INDEX_INV(VAR_2D_DT_INDEX, d_var_2d_dt);



//--------------------------------------------------------------------------------------------------
// 3D unknowns
//--------------------------------------------------------------------------------------------------
const VAR VAR_3D_INDEX[]={U, V, W

#ifdef BAROCLINIC_KEY  // Baroclinic case unknowns
                          , T, S, RHO, RHO_FRS
#endif

#ifdef MASS_CONSERVATION_KEY  // global mass conservation unknown
                          , TRACER
#endif

#if TURBULENCE_KEY>1   // "k" turbulence model unknown
#ifdef BAROCLINIC_KEY  // Baroclinic case unknowns: density gradient
                          , RHO_Z
#endif
                          , K
#if TURBULENCE_KEY==3  // "k-epsilon" and Mellor-Yamada turbulence model unknown
                          , E
#endif
#endif
};

const int d_var_3d = sizeof VAR_3D_INDEX / sizeof *VAR_3D_INDEX;

const Inverse VAR_3D_INDEX_INV(VAR_3D_INDEX, d_var_3d);



//--------------------------------------------------------------------------------------------------
// Time dependent 3D unknowns
//--------------------------------------------------------------------------------------------------
const VAR VAR_3D_DT_INDEX[]={U, V

#ifdef BAROCLINIC_KEY  // Baroclinic case unknowns
                             , T, S
#endif

#ifdef MASS_CONSERVATION_KEY  // global mass conservation unknown
, TRACER
#endif

#if TURBULENCE_KEY>1   // "k" turbulence model unknown
                             , K
#if TURBULENCE_KEY==3  // "k-epsilon" and Mellor-Yamada turbulence model unknown
                             , E
#endif
#endif
};

const int d_var_3d_dt = sizeof VAR_3D_DT_INDEX / sizeof *VAR_3D_DT_INDEX;

const Inverse VAR_3D_DT_INDEX_INV(VAR_3D_DT_INDEX, d_var_3d_dt);



//--------------------------------------------------------------------------------------------------
// Time dependent unknowns
//--------------------------------------------------------------------------------------------------
const VAR VAR_DT_INDEX[]={XI, U, V

#ifdef BAROCLINIC_KEY  // Baroclinic case unknowns
                          , T, S
#endif

#ifdef MASS_CONSERVATION_KEY  // global mass conservation unknown
, TRACER
#endif

#if TURBULENCE_KEY>1   // "k" turbulence model unknown
                          , K
#if TURBULENCE_KEY==3  // "k-epsilon" and Mellor-Yamada turbulence model unknown
                          , E
#endif
#endif
};

const int d_var_dt = sizeof VAR_DT_INDEX / sizeof *VAR_DT_INDEX;

const Inverse VAR_DT_INDEX_INV(VAR_DT_INDEX, d_var_dt);



//--------------------------------------------------------------------------------------------------
// Time independent unknowns
//--------------------------------------------------------------------------------------------------
const VAR VAR_C_INDEX[]={W

#ifdef BAROCLINIC_KEY  // Baroclinic case unknowns
                         , RHO, RHO_FRS
#if TURBULENCE_KEY>1   // turbulence model unknowns
                         , RHO_Z
#endif
#endif

#ifdef DEPTH_AVERAGED_VELOCITY_KEY  // algebraic turbulence model unknowns
                         , UH, VH
#endif
};

const int d_var_c = sizeof VAR_C_INDEX / sizeof *VAR_C_INDEX;

const Inverse VAR_C_INDEX_INV(VAR_C_INDEX, d_var_c);



//--------------------------------------------------------------------------------------------------
// Transport unknowns
//--------------------------------------------------------------------------------------------------
const VAR VAR_TRANSPORT_INDEX[]={

#ifdef BAROCLINIC_KEY  // Baroclinic case unknowns
#define COMMA_KEY
                                 T, S
#endif

#ifdef MASS_CONSERVATION_KEY  // global mass conservation unknown
#ifdef COMMA_KEY
                                 ,
#endif 
#define COMMA_KEY
                                 TRACER
#endif

#if TURBULENCE_KEY>1   // "k" turbulence model unknown
#ifdef COMMA_KEY
                                 ,
#endif 
#define COMMA_KEY
                                 K

#if TURBULENCE_KEY==3  // "k-epsilon" and Mellor-Yamada turbulence model unknown
                                 , E
#endif
#endif
#undef COMMA_KEY
};

const int d_var_transport = sizeof VAR_TRANSPORT_INDEX / sizeof *VAR_TRANSPORT_INDEX;

const Inverse VAR_TRANSPORT_INDEX_INV(VAR_TRANSPORT_INDEX, d_var_transport);
#endif
