//----------------------------------------------------------------------------------------
// mesh_2d.cpp: Contains definitions for nodes, edges, and triangles in 2D
//----------------------------------------------------------------------------------------

#include <cassert>
#include <cmath>
#include "general.H"
#include "mesh_2d.H"

//--------------------------------------------------------------------------------------
// NODE_2D Class: Node class in 2D
//--------------------------------------------------------------------------------------

// Default constructor
NODE_2D::NODE_2D()
{
  id=BAD_INDEX;
  scale=coord[0]=coord[1]=0.;
}

// Constructor for hot start input
// Parameters: hot start input file stream
NODE_2D::NODE_2D(std::ifstream& in)
{
  in.read((char*)this, sizeof *this);
}

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const NODE_2D& node)
{
  out.write((char*)&node, sizeof node);
  return out;
}



//--------------------------------------------------------------------------------------
// EDGE_2D Class: Edge class in 2D
//--------------------------------------------------------------------------------------

// Default constructor
EDGE_2D::EDGE_2D()
{
  normal[0]=normal[1]=0.;
  status=n_child=0;
  type=id=node[0]=node[1]=element[0]=element[1]=parent=child=BAD_INDEX;
}

// Constructor for hot start input
// Parameters: hot start input file stream
EDGE_2D::EDGE_2D(std::ifstream& in)
{
  in.read((char*)this, sizeof *this);
}

// Compute mapping from the reference edge
// Parameters: node_coord - node coordinates
void EDGE_2D::ComputeReferenceMapping(const real *node_coord[2])
{
  mapping.UpdateMap(node_coord);
  normal[0]=(node_coord[1][1]-node_coord[0][1])/mapping.GetScale();
  normal[1]=(node_coord[0][0]-node_coord[1][0])/mapping.GetScale();
}

// Refine the edge using edge bisection
// Parameters: edge_child - child edges; node - nodes on the edge; node_new - new node
// Return value: id of the new node
void EDGE_2D::Refine(EDGE_2D* edge_child[2], NODE_2D* node[], NODE_2D* node_new)
{
  // Create a new node
  node_new->Coord(0)=node_new->Coord(1)=0.;
  for (int i=0; i<2; i++)
    for (int j=0; j<d_space_2d; node_new->Coord(j)+=0.5*node[i]->Coord(j), j++);

  // Create new edges
  edge_child[0]->Parent()=edge_child[1]->Parent()=id;
  edge_child[0]->Element(0)=edge_child[1]->Element(0)=Element(0);
  edge_child[0]->Element(1)=edge_child[1]->Element(1)=Element(1);
  edge_child[0]->Node(0)=node[0]->ID();
  edge_child[0]->Node(1)=edge_child[1]->Node(0)=node_new->ID();
  edge_child[1]->Node(1)=node[1]->ID();
  edge_child[0]->Type()=edge_child[1]->Type()=type;
  edge_child[0]->Normal(0)=edge_child[1]->Normal(0)=normal[0];
  edge_child[0]->Normal(1)=edge_child[1]->Normal(1)=normal[1];
  edge_child[0]->Status()=edge_child[1]->Status()=status;
  SetChildren(2, edge_child[0]->ID());
  status|=SPLIT;
}

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const EDGE_2D& edge)
{
  out.write((char*)&edge, sizeof edge);
  return out;
}



//--------------------------------------------------------------------------------------
// TRIANGLE_2D Class: Element class in 2D
//--------------------------------------------------------------------------------------

// Default constructor
TRIANGLE_2D::TRIANGLE_2D()
{
  for (int i=0; i<vert_h; i++)
    node[i]=edge[i]=index[i]=BAD_INDEX;
  type=id=parent=child=BAD_INDEX;
  status=n_child=refinement_level=0;
}

// Constructor for hot start input
// Parameters: hot start input file stream
TRIANGLE_2D::TRIANGLE_2D(std::ifstream& in)
{
  in.read((char*)this, sizeof *this);
}

// Parameters: coord - point coordinates
// Return value: Distance from coord[] to the barycenter of the element
long_real TRIANGLE_2D::DistanceToBarycenter(const real coord[]) const
{
  static const real xi_bar[d_space_2d]={1./3., 1./3.};
  real xi[d_space_2d], coord_bar[d_space_2d];

  mapping.mapto_inverse(coord, xi);
  if ((xi[0]>=0.) && (xi[1]>=0.) && (xi[0]+xi[1]<=1.))
    return 0.;  // Point inside of the element
  // Determine distance to the barycenter of the element
  mapping.mapto(xi_bar, coord_bar);
  return std::sqrt((coord[0]-coord_bar[0])*(coord[0]-coord_bar[0])+
                   (coord[1]-coord_bar[1])*(coord[1]-coord_bar[1]));
}

// Parameters: coord - point coordinates
// Return value: Distance from coord[] to the boundary of the element (0 if the point is inside)
long_real TRIANGLE_2D::DistanceToBoundary(const real coord[]) const
{
  real xi[d_space_2d], xi_bound[d_space_2d]={0., 0.}, coord_bound[d_space_2d];

  mapping.mapto_inverse(coord, xi);
  if ((xi[0]>=0.) && (xi[1]>=0.) && (xi[0]+xi[1]<=1.))
    return 0.;  // Point inside of the element
  // Determine distance to the boundary of the element
  if (xi[0]>0.)
    if (xi[1]>0.)
      if ((xi[1]-xi[0])<-1.)
        xi_bound[0]=1.;
      else 
        if ((xi[1]-xi[0])<1.)
          xi_bound[0]=0.5*(1.-xi[1]+xi[0]), 
          xi_bound[1]=1.-xi_bound[0];
        else
          xi_bound[1]=1.;
    else
      if (xi[0]<1.)
        xi_bound[0]=xi[0];
      else
        xi_bound[0]=1.;
  else
    if (xi[1]>0.)
      if (xi[1]<1.)
        xi_bound[1]=xi[1];
      else
        xi_bound[1]=1.;
  mapping.mapto(xi_bound, coord_bound);
  return std::sqrt((coord[0]-coord_bound[0])*(coord[0]-coord_bound[0])+
                   (coord[1]-coord_bound[1])*(coord[1]-coord_bound[1]));
}

// Refine the triangle using edge bisection
// Parameters: element_child - child elements; edge_new - new edges
//             edge_child - child edges; node - nodes on the edge; node_new - new nodes
// Return value: id of the new node
void TRIANGLE_2D::Refine(TRIANGLE_2D* element_child[], EDGE_2D* edge_new[vert_h], 
                        EDGE_2D* edge_child[vert_h][2], NODE_2D* node_new[vert_h])
{
  const int n_children_new=4;

  // Create new elements
  for (int i=0; i<n_children_new; i++)
  {
    if(i<vert_h)
    {
      int ind_1=(i+1)%vert_h, ind_2=(i+2)%vert_h, index_loc[vert_h]={0, index[ind_1], index[ind_2]};
      int node_loc[vert_h]={node[i], node_new[ind_2]->ID(), node_new[ind_1]->ID()}, 
        node_edge[2]={node_loc[1], node_loc[2]};
      int edge_loc[vert_h]={edge_new[i]->ID(), edge_child[ind_1][1-index[ind_1]]->ID(), 
        edge_child[ind_2][index[ind_2]]->ID()};
      int element_edge[2]={element_child[i]->ID(), element_child[n_children_new-1]->ID()};

      for (int j=0; j<vert_h; j++)
      {
        element_child[i]->Node(j)=node_loc[j];
        element_child[i]->Edge(j)=edge_loc[j];
        element_child[i]->ElementIndex(j)=index_loc[j];
      }
      for (int j=0; j<2; j++)
      {
        edge_new[i]->Node(j)=node_edge[j];
        edge_new[i]->Element(j)=element_edge[j];
      }
      edge_child[ind_1][1-index[ind_1]]->Element(index[ind_1])=
        edge_child[ind_2][index[ind_2]]->Element(index[ind_2])=element_child[i]->ID();
    }
    else
    {
      int index_loc[vert_h]={1, 1, 1};
      int node_loc[vert_h]={node_new[0]->ID(), node_new[1]->ID(), node_new[2]->ID()};
      int edge_loc[vert_h]={edge_new[0]->ID(), edge_new[1]->ID(), edge_new[2]->ID()};

      for (int j=0; j<vert_h; j++)
      {
        element_child[i]->Node(j)=node_loc[j];
        element_child[i]->Edge(j)=edge_loc[j];
        element_child[i]->ElementIndex(j)=index_loc[j];
      }
    }
    element_child[i]->RefinementLevel()=RefinementLevel()+1;
    element_child[i]->Parent()=id;
    element_child[i]->Type()=type;
    element_child[i]->Status()=status;
  }
  SetChildren(n_children_new, element_child[0]->ID());
  status|=SPLIT;
}

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const TRIANGLE_2D& element)
{
  out.write((char*)&element, sizeof element);
  return out;
}
