//----------------------------------------------------------------------------------------
// mesh_2d_sw.cpp: Contains definitions for 2D node, edge, and triangle shallow-water objects
//----------------------------------------------------------------------------------------

#include <cassert>
#include <cmath>
#include <iostream>
#include <fstream>
#include "general.H"
#include "pde.H"
#include "mesh_2d_sw.H"
#include "mesh_container.H"
#ifdef TEST_KEY
#include "test.H"
#endif

extern PROBLEM_DESCRIPTION problem_info;  // Storage for physical parameters
extern QUAD_CONTAINER quadrature_index;   // Quadrature objects
extern BASIS_2D* basis_2d[max_order+1];   // 2D basis objects
extern DOF var[d_var];                    // Global storage for degrees of freedom for unknowns
extern MESH mesh;                         // Mesh object


//--------------------------------------------------------------------------------------
// NODE_2D_SW Class: Node class for shallow-water problems
//--------------------------------------------------------------------------------------

// Default constructor
NODE_2D_SW::NODE_2D_SW()
{
  xi=depth=coord_spher[0]=coord_spher[1]=0.;
}

// Constructor for hot start input
// Parameters: in - hot start input file stream
NODE_2D_SW::NODE_2D_SW(std::ifstream& in)
{
  in.read((char*)this, sizeof *this);
}

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const NODE_2D_SW& node)
{
  out.write((char*)&node, sizeof node);
  return out;
}


//--------------------------------------------------------------------------------------
// EDGE_2D_SW Class: Edge class for shallow-water problems
//--------------------------------------------------------------------------------------

// Default constructor
EDGE_2D_SW::EDGE_2D_SW() : EDGE_2D()
{
  Type()=UNKNOWN;
  xi_0=d_xi=dp_0=d_dp=0.;
  coef_scale_d[0]=coef_scale_d[1]=0.;
  n_basis[0]=n_basis[1]=0;
  for (int i=0; i<d_var; basis_index[0][i]=basis_index[1][i]=BAD_INDEX, i++);
  quad=quadrature_index.GetEdgeQuad(problem_info.def_order);
}

// Constructor that copies data from an object of EDGE_2D type
// Parameters: edge - object to be copied
EDGE_2D_SW::EDGE_2D_SW(const EDGE_2D& edge) : EDGE_2D(edge)
{
  Type()=UNKNOWN;
  xi_0=d_xi=dp_0=d_dp=0.;
  coef_scale_d[0]=coef_scale_d[1]=0.;
  n_basis[0]=n_basis[1]=0;
  for (int i=0; i<d_var; basis_index[0][i]=basis_index[1][i]=BAD_INDEX, i++);
  quad=quadrature_index.GetEdgeQuad(problem_info.def_order);
}

// Constructor for hot start input
// Parameters: in - hot start input file stream
EDGE_2D_SW::EDGE_2D_SW(std::ifstream& in) : EDGE_2D(in)
{
  in.read((char*)&xi_0, sizeof xi_0);
  in.read((char*)&d_xi, sizeof d_xi);
  in.read((char*)&dp_0, sizeof dp_0);
  in.read((char*)&d_dp, sizeof d_dp);
  in.read((char*)coef_scale_d, sizeof coef_scale_d);
  in.read((char*)n_basis, sizeof n_basis);
  in.read((char*)basis_index, sizeof basis_index);
  quad=quadrature_index.GetEdgeQuad(problem_info.def_order);
}

// Compute mapping from the reference edge
void EDGE_2D_SW::ComputeReferenceMapping()
{
  const real* node_coord[2]={mesh.node_2d_list[Node(0)].Coord(), mesh.node_2d_list[Node(1)].Coord()};

  EDGE_2D::ComputeReferenceMapping(node_coord); 
  dp_0=mesh.node_2d_list[Node(0)].Depth();
  d_dp=mesh.node_2d_list[Node(1)].Depth()-dp_0;
}

// Calculate coefficients for surface elevations computation
void EDGE_2D_SW::ComputeSurfaceElevation()
{
  xi_0=mesh.node_2d_list[Node(0)].Xi();
  d_xi=mesh.node_2d_list[Node(1)].Xi()-xi_0;
}

// Calculate coefficients for surface elevations computation
// Parameters: elev - surface elevation at the nodes
void EDGE_2D_SW::ComputeSurfaceElevation(real elev[2])
{
  xi_0=mesh.node_2d_list[Node(0)].Xi()=elev[0];
  d_xi=mesh.node_2d_list[Node(1)].Xi()=elev[1];
  d_xi-=xi_0;
}

// Update recursively edge-element connectivity info
// Parameters: el_ind - element index in the edge; element_new - element id
void EDGE_2D_SW::SetElementRecursive(int el_ind, int element_new)
{
  Element(el_ind)=element_new;
  for (int i=0; i<NChildren(); i++)
    mesh.edge_2d_list[Child(i)].SetElementRecursive(el_ind, element_new);
}

// Set recursively the edge (or its descendants) to be inactive
// This version works for bundary edges and must be overloaded for interior edges
void EDGE_2D_SW::SetActiveRecursive()
{
  assert_debug(IsElement(0));
  if (mesh.triangle_2d_list[Element(0)].IsActive())
  {
    Status()|=ACTIVE;
    UnSetUpdated();
  }
  else
  {
    Status()&=~ACTIVE;
    for (int i=0; i<NChildren(); i++)
      mesh.edge_2d_list[Child(i)].SetActiveRecursive();
  }
}

// Set recursively the edge (or its descendants) to be inactive
void EDGE_2D_SW::UnSetActiveRecursive()
{
  Status()&=~ACTIVE;
  for (int i=0; i<NChildren(); i++)
    mesh.edge_2d_list[Child(i)].UnSetActiveRecursive();
}


#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
// Refine the edge using edge bisection
// Parameters: node_new_ind - id of the new node, lvalue
// Return value: id of the first child edge
void EDGE_2D_SW::Refine(int& node_new_ind)
{
  if(!IsRefined())
  {
    const int n_children_new=2;
    real coord_tmp[d_space_2d];
    NODE_2D* node[2]={&mesh.node_2d_list[Node(0)], &mesh.node_2d_list[Node(1)]};
    EDGE_2D* edge_child[n_children_new];
    NODE_2D_SW node_new;

    // Create a new node
    node_new_ind=node_new.ID()=mesh.node_2d_list.size();
    problem_info.CartesianToSpherical(node_new.Coord(), coord_tmp);
    node_new.SpherCoord(0)=coord_tmp[0];
    node_new.SpherCoord(1)=coord_tmp[1];
    node_new.Depth()=0.;
    // Create new edges
    for (int i=0; i<n_children_new; i++)
    {
      EDGE_2D_SW* edge_child_loc=CreateEdge();

      edge_child_loc->ID()=mesh.edge_2d_list.size();
      mesh.edge_2d_list.push_back(*edge_child_loc);
      edge_child[i]=&mesh.edge_2d_list[edge_child_loc->ID()];
      delete edge_child_loc;
      node_new.Depth()+=0.5*((NODE_2D_SW*)node[i])->Depth();
    }
    mesh.node_2d_list.push_back(node_new);
    EDGE_2D::Refine(edge_child, node, &mesh.node_2d_list[node_new_ind]);
    for (int i=0; i<n_children_new; i++)
    {
      EDGE_2D_SW* ed_child=(EDGE_2D_SW*)edge_child[i];

      ed_child->ComputeReferenceMapping();
      ed_child->InitializeStorage();
    }
  }
  else
    node_new_ind=mesh.edge_2d_list[Child(0)].Node(1);
}
#endif

// Initialize storage for data stored at quadrature points
void EDGE_2D_SW::InitializeStorage()
{
  int n_quad=Quad()->GetNQuad();
  std::size_t ind=mass_storage_.size();

  BaseStorageEdgeSW::Initialize(n_quad);
  for (int k=0; k<2; k++)
    if (IsElement(k))
    {
      TRIANGLE_2D_SW* element=&mesh.triangle_2d_list[Element(k)];

      n_basis[k]=element->NBasis();
      for (int i=0; i<d_var; i++)
        basis_index[k][i]=element->Basis_index((VAR)i);
      for (int j=0; j<n_basis[k]; j++)
      {
        int n_dim=element->Basis(j)->Dim();

        mass_storage_.insert(mass_storage_.end(), n_quad*n_dim, 0);
        for (SetIndex(0); InRange(); ind+=n_dim, (*this)++)
          quad_local().basis_func_ind[k][j]=ind;
      }
    }
}

// Update data shared by all edges
// Parameters: var_new - global storage for degrees of freedom;
//             var_rhs_new - global storage for degrees of freedom for the right hand side
void EDGE_2D_SW::Update_static(DOF var_new[d_var], DOF var_rhs_new[d_var])
{
  for (int i=0; i<d_var; i++)
  {
    temp_storage.Var_dof[i]=var_new[i].GetDofIterator();
    temp_storage.L_Var_dof[i]=var_rhs_new[i].GetDofIterator();
  }
}

// Update rarely changed data at quadrature points (basis functions, coordinates, etc)
void EDGE_2D_SW::Update_basis_quad()
{
  const long_real direction_coef[2]={1., -1.};
  assert_debug(IsInitialized());
  for (SetIndex(0); InRange(); (*this)++)
  {
    Mapping().mapto(Quad()->p(GetIndex()), coord_quad());
    depth_quad()=Depth(quad->p(GetIndex()));
  }
  for (int k=0; k<2; k++)
    if (IsElement(k))
    {
      TRIANGLE_2D_SW& element=mesh.triangle_2d_list[Element(k)];

      coef_scale(k)=direction_coef[k]/element.Mapping().GetScale();
      for (SetIndex(0); InRange(); (*this)++)
        for (int j=0; j<n_basis[k]; j++)
          element.Basis(j)->
            basis_func_phys(coord_quad(), element.Mapping(), basis_func_assign_quad(k, j));
    }
}

// Update often changed data at quadrature points (unknowns)
// Parameters: t - time
void EDGE_2D_SW::Update_variables_quad(real t)
{
  assert_debug(IsUpdated());
  for (SetIndex(0); InRange(); (*this)++)
    for (int k=0; k<2; k++)
      if (IsElement(k))
        quad_local().Var[k][VAR_2D_DT_INDEX_INV[XI]]=
          mesh.triangle_2d_list[Element(k)].Var_dof(XI)*basis_func_quad(k, XI);
      else
      {
        real var[d_var], param[1];

        BC_Dirichlet_quad(t, param, var);
        quad_local().Var[k][VAR_2D_DT_INDEX_INV[XI]]=var[XI];
      }
}

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const EDGE_2D_SW& edge)
{
  out << (const EDGE_2D)edge;
  out.write((char*)&edge.xi_0, sizeof edge.xi_0);
  out.write((char*)&edge.d_xi, sizeof edge.d_xi);
  out.write((char*)&edge.dp_0, sizeof edge.dp_0);
  out.write((char*)&edge.d_dp, sizeof edge.d_dp);
  out.write((char*)edge.coef_scale_d, sizeof edge.coef_scale_d);
  out.write((char*)edge.n_basis, sizeof edge.n_basis);
  out.write((char*)edge.basis_index, sizeof edge.basis_index);
  return out;
}



//--------------------------------------------------------------------------------------
// EDGE_INTERIOR Class: Interior edge class for shallow-water problems
//--------------------------------------------------------------------------------------

// Default constructor
EDGE_INTERIOR::EDGE_INTERIOR() : EDGE_2D_SW()
{
  Type()=INTERIOR;
}

// Constructor
EDGE_INTERIOR::EDGE_INTERIOR(const EDGE_2D_SW& edge) : EDGE_2D_SW(edge)
{
  Type()=INTERIOR;
}

// Set recursively the edge (or its descendants) to be active
void EDGE_INTERIOR::SetActiveRecursive()
{
  if (IsElement(0) && mesh.triangle_2d_list[Element(0)].IsActive() && 
    IsElement(1) && mesh.triangle_2d_list[Element(1)].IsActive())
  {
    Status()|=ACTIVE;
    UnSetUpdated();
  }
  else
  {
    Status()&=~ACTIVE;
    for (int i=0; i<NChildren(); i++)
      mesh.edge_2d_list[Child(i)].SetActiveRecursive();
  }
}



//--------------------------------------------------------------------------------------
// EDGE_LAND Class: Land edge class for shallow-water problems
//--------------------------------------------------------------------------------------

// Default constructor
EDGE_LAND::EDGE_LAND() : EDGE_2D_SW()
{
  Type()=LAND;
}

// Constructor
EDGE_LAND::EDGE_LAND(const EDGE_2D_SW& edge) : EDGE_2D_SW(edge)
{
  Type()=LAND;
}



//--------------------------------------------------------------------------------------
// EDGE_SEA Class: Sea edge class for shallow-water problems
//--------------------------------------------------------------------------------------

// Default constructor
EDGE_SEA::EDGE_SEA() : EDGE_2D_SW()
{
  Type()=SEA;
  n_freq=problem_info.n_bound_tp_freq;
  std::size_t ind=mass_storage_.size();
  mass_storage_.insert(mass_storage_.end(), 4*n_freq, (real)0.);
  amplitude_ind=ind;
  phase_ind=ind+n_freq;
  tidal_coef_ind[0]=ind+2*n_freq;
  tidal_coef_ind[1]=ind+3*n_freq;
  tidal_elevation=0.;
#ifdef BAROCLINIC_KEY
  temperature=problem_info.T_0;
  salinity=problem_info.S_0;
#endif
}

// Constructor
EDGE_SEA::EDGE_SEA(const EDGE_2D_SW& edge) : EDGE_2D_SW(edge)
{
  Type()=SEA;
  n_freq=problem_info.n_bound_tp_freq;
  std::size_t ind=mass_storage_.size();
  mass_storage_.insert(mass_storage_.end(), 4*n_freq, (real)0.);
  amplitude_ind=ind;
  phase_ind=ind+n_freq;
  tidal_coef_ind[0]=ind+2*n_freq;
  tidal_coef_ind[1]=ind+3*n_freq;
  tidal_elevation=0.;
#ifdef BAROCLINIC_KEY
  temperature=problem_info.T_0;
  salinity=problem_info.S_0;
#endif
}

// Constructor for hot start input
// Parameters: in - hot start input file stream
EDGE_SEA::EDGE_SEA(std::ifstream& in) : EDGE_2D_SW(in)
{
  std::size_t ind=mass_storage_.size();

  in.read((char*)&n_freq, sizeof(n_freq));
  in.read((char*)&tidal_elevation, sizeof(tidal_elevation));
#ifdef BAROCLINIC_KEY
  in.read((char*)&temperature, sizeof(temperature));
  in.read((char*)&salinity, sizeof(salinity));
#endif
  mass_storage_.insert(mass_storage_.end(), 4*n_freq, (real)0.);
  amplitude_ind=ind;
  phase_ind=ind+n_freq;
  tidal_coef_ind[0]=ind+2*n_freq;
  tidal_coef_ind[1]=ind+3*n_freq;
  if (n_freq>0)
  {
    in.read((char*)Amplitude(), n_freq*sizeof(*Amplitude()));
    in.read((char*)Phase(), n_freq*sizeof(*Phase()));
    for (int i=0; i<n_freq; i++)
      ComputeTidalCoefficients(i);
  }
}

// Compute coefficients for tidal forcing
// Parameters: i - forcing component index
void EDGE_SEA::ComputeTidalCoefficients(int i)
{
  TidalCoef(i, 0)=Amplitude(i)*problem_info.bound_tp_param[i].ff*
    std::cos(problem_info.bound_tp_param[i].phase-Phase(i));
  TidalCoef(i, 1)=Amplitude(i)*problem_info.bound_tp_param[i].ff*
    std::sin(problem_info.bound_tp_param[i].phase-Phase(i));
}

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
// Refine the edge using edge bisection
// Parameters: node_new_ind - id of the new node, lvalue
// Return value: id of the first child edge
void EDGE_SEA::Refine(int& node_new_ind)
{
  if(IsRefined())
    return EDGE_2D_SW::Refine(node_new_ind);
  else
  {
    EDGE_2D_SW::Refine(node_new_ind);
    for (int i=0; i<NChildren(); i++)
    {
      EDGE_SEA& edge_child=(EDGE_SEA&)mesh.edge_2d_list[Child(i)];

      for (int j=0; j<n_freq; j++)
      {
        edge_child.Amplitude(j)=Amplitude(j);
        edge_child.Phase(j)=Phase(j);
        edge_child.ComputeTidalCoefficients(j);
      }
      edge_child.TidalElevation()=TidalElevation();
#ifdef BAROCLINIC_KEY
      edge_child.Temperature()=Temperature();
      edge_child.Salinity()=Salinity();
#endif
    }
  }
}
#endif

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const EDGE_SEA& edge)
{
  out << (const EDGE_2D_SW)edge;
  out.write((char*)&edge.n_freq, sizeof edge.n_freq);
  out.write((char*)&edge.tidal_elevation, sizeof edge.tidal_elevation);
#ifdef BAROCLINIC_KEY
  out.write((char*)&edge.temperature, sizeof edge.temperature);
  out.write((char*)&edge.salinity, sizeof edge.salinity);
#endif
  if (edge.n_freq>0)
  {
    out.write((char*)edge.Amplitude(), edge.n_freq*sizeof(*edge.Amplitude()));
    out.write((char*)edge.Phase(), edge.n_freq*sizeof(*edge.Phase()));
  }
  return out;
}



//--------------------------------------------------------------------------------------
// EDGE_RIVER Class: River edge class for shallow-water problems
//--------------------------------------------------------------------------------------

// Default constructor
EDGE_RIVER::EDGE_RIVER() : EDGE_2D_SW()
{
  Type()=RIVER;
  elevation=velocity[0]=velocity[1]=0.;
#ifdef BAROCLINIC_KEY
  temperature=problem_info.T_0;
  salinity=problem_info.S_0;
#endif
}

// Constructor
EDGE_RIVER::EDGE_RIVER(const EDGE_2D_SW& edge) : EDGE_2D_SW(edge)
{
  Type()=RIVER;
  elevation=velocity[0]=velocity[1]=0.;
#ifdef BAROCLINIC_KEY
  temperature=problem_info.T_0;
  salinity=problem_info.S_0;
#endif
}

// Constructor for hot start input
// Parameters: in - hot start input file stream
EDGE_RIVER::EDGE_RIVER(std::ifstream& in) : EDGE_2D_SW(in)
{
  in.read((char*)&elevation, sizeof(elevation));
  in.read((char*)velocity, sizeof(velocity));
#ifdef BAROCLINIC_KEY
  in.read((char*)&temperature, sizeof(temperature));
  in.read((char*)&salinity, sizeof(salinity));
#endif
}

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
// Refine the edge using edge bisection
// Parameters: node_new_ind - id of the new node, lvalue
// Return value: id of the first child edge
void EDGE_RIVER::Refine(int& node_new_ind)
{
  if(IsRefined())
    return EDGE_2D_SW::Refine(node_new_ind);
  else
  {
    EDGE_2D_SW::Refine(node_new_ind);
    for (int i=0; i<NChildren(); i++)
    {
      EDGE_RIVER& edge_child=(EDGE_RIVER&)mesh.edge_2d_list[Child(i)];

      edge_child.SurfaceElevation()=SurfaceElevation();
      for (int j=0; j<d_space_2d; j++)
        edge_child.Velocity(j)=Velocity(j);
#ifdef BAROCLINIC_KEY
      edge_child.Temperature()=Temperature();
      edge_child.Salinity()=Salinity();
#endif
    }
  }
}
#endif

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const EDGE_RIVER& edge)
{
  out << (const EDGE_2D_SW)edge;
  out.write((char*)&edge.elevation, sizeof(edge.elevation));
  out.write((char*)edge.velocity, sizeof(edge.velocity));
#ifdef BAROCLINIC_KEY
  out.write((char*)&edge.temperature, sizeof(edge.temperature));
  out.write((char*)&edge.salinity, sizeof(edge.salinity));
#endif
  return out;
}



//--------------------------------------------------------------------------------------
// EDGE_RADIATION Class: Radiation edge class for shallow-water problems
//--------------------------------------------------------------------------------------

// Default constructor
EDGE_RADIATION::EDGE_RADIATION() : EDGE_2D_SW()
{
  Type()=RADIATION;
}

// Constructor
EDGE_RADIATION::EDGE_RADIATION(const EDGE_2D_SW& edge) : EDGE_2D_SW(edge)
{
  Type()=RADIATION;
}



#ifdef TEST_KEY
//--------------------------------------------------------------------------------------
// EDGE_UNKNOWN Class: Test (Dirichlet) edge class for shallow-water problems
//--------------------------------------------------------------------------------------

// Default constructor
EDGE_UNKNOWN::EDGE_UNKNOWN() : EDGE_2D_SW()
{
  Type()=UNKNOWN;
}

// Constructor
EDGE_UNKNOWN::EDGE_UNKNOWN(const EDGE_2D_SW& edge) : EDGE_2D_SW(edge)
{
  Type()=UNKNOWN;
}
#endif



//--------------------------------------------------------------------------------------
// TRIANGLE_2D_SW Class: Element class for shallow-water problems
//--------------------------------------------------------------------------------------

// Default constructor
TRIANGLE_2D_SW::TRIANGLE_2D_SW()
{
  static const VAR* ind=VAR_2D_INDEX;

  order=problem_info.def_order;
  n_basis=0;
  xi_0=d_xi[0]=d_xi[1]=dp_0=d_dp[0]=d_dp[1]=0.;
  tidal_forcing[0]=tidal_forcing[1]=0.;
  for (int i=0; i<d_var; i++)
    basis_index[i]=BAD_INDEX;

  // Initialize pointers to basis objects
  for (int i=0; i<d_var_2d; i++)
  {
    for (int j=0; j<i; j++)
      if (problem_info.var_order[ind[i]]==problem_info.var_order[ind[j]])
      {
        basis_index[ind[i]]=basis_index[ind[j]];
        break;
      }
    // Assign new basis object
    if (basis_index[ind[i]]==BAD_INDEX)
    {
      basis[n_basis]=basis_2d[problem_info.var_order[ind[i]]];
      basis_index[ind[i]]=n_basis++;
    }
  }
  for (int i=n_basis; i<d_var_2d; i++)
    basis[i]=0;

//   // Assign basis functions and index variables to them
//   n_basis=Order()+1;
//   for (int i=0; i<d_var_2d; i++)
//     basis[i] = i<n_basis ? basis_2d[Order()-i] : 0;
//   for (int i=0; i<d_var; i++)
//   {
//     if (VAR_2D_INDEX_INV[i]==BAD_INDEX)
//     {
//       basis_index[i]=BAD_INDEX;
//       continue;
//     }
//     basis_index[i]=VAR_BASIS_INDEX[i];
//   }
  quad=quadrature_index.GetTriangleQuad(order);
  vert=quadrature_index.GetTriangleNodeQuad();
  InitializeStorage();
#ifdef ADAPTIVITY_H_2D_KEY
  min=max=abs_max=s_jump[0]=s_jump[1]=s_boundary[0]=s_boundary[1]=0.;
#endif
  if ((problem_info.tidal_poten)&&(problem_info.n_tp_freq>0))
  {
    tidal_coef_ind=mass_storage_.size();
    mass_storage_.insert(mass_storage_.end(), 4*problem_info.n_tp_freq, (real)0.);
  }
  else
    tidal_coef_ind=(std::size_t)BAD_INDEX;
}

// Constructor for hot start input
// Parameters: in - hot start input file stream
TRIANGLE_2D_SW::TRIANGLE_2D_SW(std::ifstream& in) : TRIANGLE_2D(in)
{
  static const VAR* ind=VAR_2D_INDEX;

  in.read((char*)&order, sizeof order);
  in.read((char*)&xi_0, sizeof xi_0);
  in.read((char*)d_xi, sizeof d_xi);
  in.read((char*)&dp_0, sizeof dp_0);
  in.read((char*)d_dp, sizeof d_dp);
  in.read((char*)&n_basis, sizeof n_basis);
  in.read((char*)basis_index, sizeof basis_index);
  tidal_forcing[0]=tidal_forcing[1]=0.;
  // Assign basis functions
  for (int i=0; i<d_var_2d; i++)
  {
    assert_debug(basis_index[ind[i]]!=BAD_INDEX);
    basis[basis_index[ind[i]]]=basis_2d[problem_info.var_order[ind[i]]];
  }
//   n_basis=Order()+1;
//   for (int i=0; i<d_var_2d; i++)
//     basis[i] = i<n_basis ? basis_2d[Order()-i] : 0;
  quad=quadrature_index.GetTriangleQuad(order);
  vert=quadrature_index.GetTriangleNodeQuad();
  if ((problem_info.tidal_poten)&&(problem_info.n_tp_freq>0))
  {
    tidal_coef_ind=mass_storage_.size();

    mass_storage_.insert(mass_storage_.end(), 4*problem_info.n_tp_freq, (real)0.);
    in.read((char*)&mass_storage(tidal_coef_ind), 4*problem_info.n_tp_freq*sizeof(real));
  }
  else
    tidal_coef_ind=(std::size_t)BAD_INDEX;
#ifdef ADAPTIVITY_H_2D_KEY
  in.read((char*)&min, sizeof min);
  in.read((char*)&max, sizeof max);
  in.read((char*)s_jump, sizeof s_jump);
  in.read((char*)s_boundary, sizeof s_boundary);
#endif
}

// Calculate coefficients for surface elevations computation
void TRIANGLE_2D_SW::ComputeSurfaceElevation()
{
  xi_0=mesh.node_2d_list[Node(0)].Xi();
  d_xi[0]=mesh.node_2d_list[Node(1)].Xi()-xi_0;
  d_xi[1]=mesh.node_2d_list[Node(2)].Xi()-xi_0; 
}

// Compute mapping from the reference triangle
void TRIANGLE_2D_SW::ComputeReferenceMapping()
{
  NODE_2D_SW node[vert_h]={mesh.node_2d_list[Node(0)], mesh.node_2d_list[Node(1)], mesh.node_2d_list[Node(2)]};
  real node_depth[vert_h]={node[0].Depth(), node[1].Depth(), node[2].Depth()};
  const real* node_coord[vert_h]={node[0].Coord(), node[1].Coord(), node[2].Coord()};

  TRIANGLE_2D::ComputeReferenceMapping(node_coord);
  ComputeDepth(node_depth);

  // Compute coefficients for the tidal potential term 
  if ((problem_info.tidal_poten)&&(problem_info.n_tp_freq>0))
    for (int i=0; i<problem_info.n_tp_freq; i++)
    {
      real grad[d_space_2d];
      long_real cos_node[vert_h], sin_node[vert_h], coef=problem_info.tp_param[i].etrf*
        problem_info.tp_param[i].tp_coef*problem_info.tp_param[i].ff;

      switch((int)(0.00014/problem_info.tp_param[i].amig+0.5))
      {
      case 1:
        for (int j=0; j<vert_h; j++)
        {
          long_real L_i=std::cos(node[j].SpherCoord(1));
          cos_node[j]=coef*L_i*L_i*std::cos(2.*node[j].SpherCoord(0)+problem_info.tp_param[i].phase);
          sin_node[j]=coef*L_i*L_i*std::sin(2.*node[j].SpherCoord(0)+problem_info.tp_param[i].phase);
        }
        break;
      case 2:
        for (int j=0; j<vert_h; j++)
        {
          long_real L_i=std::sin(2.*node[j].SpherCoord(1));
          cos_node[j]=coef*L_i*std::cos(node[j].SpherCoord(0)+problem_info.tp_param[i].phase);
          sin_node[j]=coef*L_i*std::sin(node[j].SpherCoord(0)+problem_info.tp_param[i].phase);
        }
        break;
      default: std::cerr << "File mesh_2d_sw.cpp: Tidal forcing frequency is not (semi-)diurnal!" << std::endl;
        std::exit(1);
      }
      grad[0]=cos_node[1]-cos_node[0];
      grad[1]=cos_node[2]-cos_node[0];
      Mapping().MapGradient(grad, &TidalCoef(i, 0, 0));
      grad[0]=sin_node[1]-sin_node[0];
      grad[1]=sin_node[2]-sin_node[0];
      Mapping().MapGradient(grad, &TidalCoef(i, 1, 0));
    }
}

// Check minimum depth condition and adjust surface elevation if necessary
// Parameters: min_depth - minimum depth
void TRIANGLE_2D_SW::MinDepthCheck(real min_depth)
{
  real delta_h=0.;

  for (int i=0; i<vert_h; i++) // Compute the minimum elevation at the vertices
  {
    long_real h_tmp=min_depth-Depth(vert->p(i))-Var_dof(XI)*basis_func_node(XI, i);
    delta_h = h_tmp>delta_h ? h_tmp : delta_h;
  }
  if (delta_h>0.) // Adjust elevation if necessary
    for (SetIndex(0); InRange(); (*this)++)
      Var_dof(XI).Add(weight_quad()*delta_h, basis_func_quad(XI));
}

#if defined(ADAPTIVITY_H_2D_KEY) || defined(ADAPTIVITY_H_3D_KEY)
// Refine the element using edge bisection
// Return value: id of the first child element
void TRIANGLE_2D_SW::Refine()
{
  if (!IsRefined()) // Perform refinement
  {
    const int n_children_new=4;
    NODE_2D *node_new[vert_h];
    EDGE_2D *edge_new[vert_h], *edge_child[vert_h][2];
    TRIANGLE_2D *element_child[n_children_new];

    // Create new edges and nodes
    for (int i=0; i<vert_h; i++)
    {
      int node_new_ind=0;
      EDGE_INTERIOR edge_new_loc;

      mesh.edge_2d_list[Edge(i)].Refine(node_new_ind);
      edge_new_loc.ID()=mesh.edge_2d_list.size();
      mesh.edge_2d_list.push_back(edge_new_loc);
      node_new[i]=&mesh.node_2d_list[node_new_ind];
      edge_child[i][0]=&mesh.edge_2d_list[mesh.edge_2d_list[Edge(i)].Child(0)];
      edge_child[i][1]=&mesh.edge_2d_list[mesh.edge_2d_list[Edge(i)].Child(1)];
      edge_new[i]=&mesh.edge_2d_list[edge_new_loc.ID()];
    }

    // Create new elements
    for (int i=0; i<n_children_new; i++)
    {
      TRIANGLE_2D_SW element_child_loc;

      element_child_loc.ID()=mesh.triangle_2d_list.size();
      mesh.triangle_2d_list.push_back(element_child_loc);
      element_child[i]=&mesh.triangle_2d_list[element_child_loc.ID()];
    }
    TRIANGLE_2D::Refine(element_child, edge_new, edge_child, node_new);
    for (int i=0; i<n_children_new; i++)
    {
      TRIANGLE_2D_SW& element_child_loc=(TRIANGLE_2D_SW&)*element_child[i];
      element_child_loc.ComputeReferenceMapping();
      element_child_loc.Update_basis_quad();
      for (int j=0; j<d_var_2d; j++)
        (Var_dof(VAR_2D_INDEX[j]).GetDof())->
        InitializeLocal(element_child_loc.ID(), Var_dof(VAR_2D_INDEX[j]).GetDim());
//         var[VAR_2D_INDEX[j]].
//         InitializeLocal(element_child_loc.ID(), Var_dof(VAR_2D_INDEX[j]).GetDim());
    }
    for (int i=0; i<vert_h; i++)
    {
      ((EDGE_INTERIOR*)edge_new[i])->ComputeReferenceMapping();
      ((EDGE_INTERIOR*)edge_new[i])->InitializeStorage();
    }
  }
  // Activate the children and project variables to them
  UnSetActive();
  for (int i=0; i<vert_h; i++)
    mesh.edge_2d_list[Edge(i)].UnSetActiveRecursive();

  for (int i=0; i<NChildren(); i++)
  {
    TRIANGLE_2D_SW& element_child=mesh.triangle_2d_list[Child(i)];

    element_child.SetActive();
    for (int j=0; j<vert_h; j++)
    {
      EDGE_2D_SW& edge=mesh.edge_2d_list[element_child.Edge(j)];

      edge.SetElementRecursive(element_child.ElementIndex(j), Child(i));
      edge.SetActiveRecursive();
    }

    // Project variables to the sub-element
    for (int j=0; j<d_var_2d; j++)
      element_child.Var_dof(VAR_2D_INDEX[j])=0.;
    for (element_child.SetIndex(0); element_child.InRange(); element_child++)
    {
      func_type basis_func[d_var_2d];
      for (int j=0; j<n_basis; j++)
        Basis(j)->basis_func_phys(element_child.coord_quad(), Mapping(), basis_func[j]);
      for (int j=0; j<d_var_2d; j++)
      {
        VAR ind=VAR_2D_INDEX[j];
        long_real var_tmp=Var_dof(ind)*basis_func[Basis_index(ind)];
        element_child.Var_dof(ind).
          Add(element_child.weight_quad()*var_tmp, element_child.basis_func_quad(ind));
      }
    }
  }      
}

// Unrefine the element
void TRIANGLE_2D_SW::UnRefine()
{
  assert_debug(NChildren()>0);
  long_real w_coef=1./(real)NChildren();

  SetActive();
  // Project variables to the parent element
  for (int i=0; i<d_var_2d; i++)
    Var_dof(VAR_2D_INDEX[i])=0.;
  for (int i=0; i<NChildren(); i++)
  {
    TRIANGLE_2D_SW& element_child=mesh.triangle_2d_list[Child(i)];

    element_child.UnSetActive();
    for (int j=0; j<vert_h; j++)
      mesh.edge_2d_list[element_child.Edge(j)].UnSetActiveRecursive();
    for (element_child.SetIndex(0); element_child.InRange(); element_child++)
    {
      func_type basis_func[d_var_2d];

      for (int j=0; j<n_basis; j++)
        Basis(j)->basis_func_phys(element_child.coord_quad(), Mapping(), basis_func[j]);
      for (int j=0; j<d_var_2d; j++)
      {
        VAR ind=VAR_2D_INDEX[j];
        long_real var_tmp=element_child.Var_dof(ind)*element_child.basis_func_quad(ind);
        Var_dof(ind).
          Add(w_coef*element_child.weight_quad()*var_tmp, basis_func[Basis_index(ind)]);
      }
    }
  }
  // Activate the edges
  for (int j=0; j<vert_h; j++)
  {
    mesh.edge_2d_list[Edge(j)].SetElementRecursive(ElementIndex(j), ID());
    mesh.edge_2d_list[Edge(j)].SetActiveRecursive();
  }
}
#endif

#ifdef ADAPTIVITY_H_2D_KEY
// Initialize data used for shock-detector
void TRIANGLE_2D_SW::InitializeAdaptivityData()
{
  for (int k=0; k<2; k++)
    s_jump[k]=s_boundary[k]=0.;
  // Compute minimum/maximum values of the adaptivity variable
  min=1.e25;
  max=-1.e25;  
  for (SetIndex(0); InRange(); (*this)++)
  {
    real z_s=Var_quad(problem_info.refinement_var);
    min = min > z_s ? z_s : min;
    max = max < z_s ? z_s : max;
  }
  abs_max = max>-min ? max : min;
}
#endif

// Initialize storage for data stroed at quadrature points
void TRIANGLE_2D_SW::InitializeStorage()
{
  int n_quad=quad->GetNQuad();
  std::size_t ind=mass_storage_.size();

  BaseStorageTriangleSW::Initialize(n_quad);
  for (int j=0; j<n_basis; j++)
  {
    int n_dim=basis[j]->Dim();

    mass_storage_.insert(mass_storage_.end(), n_quad*d_space_2d*n_dim, 0);
    for (SetIndex(0); InRange(); (*this)++)
      for (int k=0; k<d_space_2d; k++)
        quad_local().basis_grad_ind[j][k]=ind, ind+=n_dim;
  }
}

// Update data shared by all edges
// Parameters: quad - pointer to quadrature object;
//             vert - pointer to quadrature object for vertices;
//             basis - pointer to basis object
//             var_new - global storage for degrees of freedom;
//             var_rhs_new - global storage for degrees of freedom for the right hand side
void TRIANGLE_2D_SW::Update_static(const QUADRATURE *quad, const QUADRATURE *vert,
                                   const BASIS *basis[max_order+1], DOF var_new[d_var],
                                   DOF var_rhs_new[d_var])
{
  static const VAR* ind=VAR_2D_INDEX;
  int i, j, n_basis=0, n_quad=quad->GetNQuad();

  for (i=0; i<n_quad; i++)  // Update weights
    weight_quad(i)=quad->w(i);

  // Initialize pointers to basis objects
  for (i=0; i<d_var_2d; i++)
  {
    for (j=0; j<i; j++)
      if (problem_info.var_order[ind[i]]==problem_info.var_order[ind[j]])
        break;
    if (i>0 && j<i)
      continue;
    // Assign new basis storage
    int order=problem_info.var_order[ind[i]], n_dim=basis[order]->Dim(), ind=mass_storage_.size();
    // Update basis functions at quadrature points
    mass_storage_.insert(mass_storage_.end(), n_quad*n_dim, 0);
    for (j=0; j<n_quad; ind+=n_dim, j++)
    {
      quad_temp_storage[j].basis_func_ind[n_basis]=ind;
      basis[order]->basis_func_ref(quad->p(j), &mass_storage_[ind]);
    }
    // Update basis functions at nodes
    mass_storage_.insert(mass_storage_.end(), vert_h*n_dim, 0);
    for (j=0; j<vert_h; ind+=n_dim, j++)
    {
      temp_storage.basis_func_node_ind[n_basis][j]=ind;
      basis[order]->basis_func_ref(vert->p(j), &mass_storage_[ind]);
    }
    n_basis++;
  }
//   for (int j=0; j<=max_order; j++)
//     if (basis[j]!=0)
//     {
//       int n_dim=basis[j]->Dim(), ind=mass_storage_.size();
//       // Update basis functions at quadrature points
//       mass_storage_.insert(mass_storage_.end(), n_quad*n_dim, 0);
//       for (int i=0; i<n_quad; ind+=n_dim, i++)
//       {
//         quad_temp_storage[i].basis_func_ind[j]=ind;
//         basis[j]->basis_func_ref(quad->p(i), &mass_storage_[ind]);
//       }
//       // Update basis functions at nodes
//       mass_storage_.insert(mass_storage_.end(), vert_h*length_basis, 0);
//       for (int i=0; i<vert_h; ind+=length_basis, i++)
//       {
//         temp_storage.basis_func_node_ind[j][i]=ind;
//         basis[j]->basis_func_ref(vert->p(i), &mass_storage_[ind]);
//       }
//     }
  // Assign pointers to global storage for degrees of freedom
  for (i=0; i<d_var; i++)
  {
    temp_storage.Var_dof[i]=var_new[i].GetDofIterator();
    temp_storage.L_Var_dof[i]=var_rhs_new[i].GetDofIterator();
  }
}

// Update rarely changed data at quadrature points (basis functions, coordinates, etc)
void TRIANGLE_2D_SW::Update_basis_quad()
{
  assert_debug(IsInitialized());
  for (SetIndex(0); InRange(); (*this)++)
  {
    const real *xi=quad->p(GetIndex());

    Mapping().mapto(xi, coord_quad());
    depth_quad()=Depth(xi);
    for (int i=0; i<n_basis; i++)
    {
      int n_dim=Basis(i)->Dim();
      grad_type basis_grad_tmp;

      Basis(i)->basis_grad_ref(xi, Mapping(), basis_grad_tmp);
      for (int k=0; k<n_dim; k++)
        for (int j=0; j<d_space_2d; j++)
          basis_grad_assign_quad(i, j)[k]=basis_grad_tmp[k][j];
    }

    // Compute Coriolis coefficient
    Coriolis_quad() = problem_info.coriolis_par==problem_info.CONSTANT ? 
      problem_info.coriolis_coef : 2.*7.29212e-5*std::sin(coord_quad(1)/problem_info.Earth_radius);
  }
}

// Update often changed data at quadrature points (unknowns)
// Parameters: t - time
void TRIANGLE_2D_SW::Update_variables_quad(real t)
{
  assert_debug(IsUpdated());
  for (SetIndex(0); InRange(); (*this)++)
    quad_local().Var[VAR_2D_DT_INDEX_INV[XI]]=Var_dof(XI)*basis_func_quad(XI);
}

// Hot start output
std::ofstream& operator<<(std::ofstream& out, const TRIANGLE_2D_SW& element)
{
  out << (const TRIANGLE_2D)element;
  out.write((char*)&element.order, sizeof element.order);
  out.write((char*)&element.xi_0, sizeof element.xi_0);
  out.write((char*)element.d_xi, sizeof element.d_xi);
  out.write((char*)&element.dp_0, sizeof element.dp_0);
  out.write((char*)element.d_dp, sizeof element.d_dp);
  out.write((char*)&element.n_basis, sizeof element.n_basis);
  out.write((char*)element.basis_index, sizeof element.basis_index);
  if ((problem_info.tidal_poten)&&(problem_info.n_tp_freq>0))
    out.write((char*)&element.mass_storage(element.tidal_coef_ind), 
              4*problem_info.n_tp_freq*sizeof(real));
#ifdef ADAPTIVITY_H_2D_KEY
  out.write((char*)&element.min, sizeof element.min);
  out.write((char*)&element.max, sizeof element.max);
  out.write((char*)element.s_jump, sizeof element.s_jump);
  out.write((char*)element.s_boundary, sizeof element.s_boundary);
#endif
  return out;
}
