//----------------------------------------------------------------------------------------
// quadrature.H: Declaration for the class used to store quadrature information
//----------------------------------------------------------------------------------------

#ifndef QUADRATURE_H
#define QUADRATURE_H 

#include <vector>
#include <cassert>
#include "general.H"


//----------------------------------------------------------------------------------------
// QUADRATURE class: Storage for quadrature information
//----------------------------------------------------------------------------------------

class QUADRATURE
{
private:
  int space_dim;                  // Dimension of the reference space
  int n_quad, n_quad_sub;         // Number of quadrature poins
  real point[max_quad][d_space];  // Coordinates of quadrature points
  long_real weight[max_quad];     // Quadrature weights

public:
  // Constructor for non tensor-product quadrature
  // Parameters: n_points - number of quadrature points; space - space dimension;
  //             p - coordinates of quadrature points; w - quadrature weights
  QUADRATURE(int n_points, int space, const real p[][d_space], const long_real w[]);

  // Constructor for tensor-product quadrature
  // Parameters: q1 - main quadrature; q2 - secondary quadrature
  QUADRATURE(const QUADRATURE* q1, const QUADRATURE* q2);

  // Default destructor
  ~QUADRATURE() {};

  // Parameters: i - quadrature point index
  // Return value: weight
  long_real w(int i) const
  {
    assert_debug(i>=0 && i<n_quad);
    return weight[i];
  }

  // Parameters: i - quadrature point index
  // Return value: quadrature point coordinates
  const real* p(int i) const
  {
    assert_debug(i>=0 && i<n_quad);
    return point[i];
  }

  // Return value: number of quadrature points
  int GetNQuad() const {return n_quad;}

  // Tensor-product quadrature only: Get index of the quadrature point in sub-quadratures
  // Parameters: n - index of sub-quadrature (0 - main, 1 - secondary), ind - quadrature point index
  // Return value: index in the corresponding sub-quadrature
  int GetSubIndex(int n, int ind) const
  {
    assert_debug(n_quad_sub!=0 && n>=0 && n<2);
    return n==0 ? ind/n_quad_sub : ind%n_quad_sub;
  }
};

//----------------------------------------------------------------------------------------
// QUAD_CONTAINER class: Storage for quadrature information for all quadrature objects
//----------------------------------------------------------------------------------------

class QUAD_CONTAINER
{
private:
  std::vector<QUADRATURE> storage;          // Main storage array
  std::size_t edge_quad[max_order+1];
  std::size_t triangle_quad[max_order+1];
  std::size_t triangle_node_quad;
  std::size_t prism_quad[max_order+1];
  std::size_t horiz_face_quad[max_order+1];
  std::size_t vert_face_quad[max_order+1];
  std::size_t prism_node_quad;

public:
  // Default constructor
  QUAD_CONTAINER();

  // Default destructor
  ~QUAD_CONTAINER(){};

  // Parameters: order - order of the quadrature rule
  // Return value: pointer to the corresponding edge quadrature object
  QUADRATURE* GetEdgeQuad(int order)
  {
    assert_debug(order>=0 && order<=max_order);
    return &storage[edge_quad[order]];
  }

  // Parameters: order - order of the quadrature rule
  // Return value: pointer to the corresponding triangle quadrature object
  QUADRATURE* GetTriangleQuad(int order)
  {
    assert_debug(order>=0 && order<=max_order);
    return &storage[triangle_quad[order]];
  }

  // Parameters: order - order of the quadrature rule
  // Return value: pointer to the corresponding triangle node quadrature object
  QUADRATURE* GetTriangleNodeQuad() {return &storage[triangle_node_quad];}

  // Parameters: order - order of the quadrature rule
  // Return value: pointer to the corresponding prism quadrature object
  QUADRATURE* GetPrismQuad(int order)
  {
    assert_debug(order>=0 && order<=max_order);
    return &storage[prism_quad[order]];
  }

  // Parameters: order - order of the quadrature rule
  // Return value: pointer to the corresponding horizontal face quadrature object
  QUADRATURE* GetHorizFaceQuad(int order)
  {
    assert_debug(order>=0 && order<=max_order);
    return &storage[triangle_quad[order]];
  }

  // Parameters: order - order of the quadrature rule
  // Return value: pointer to the corresponding vertical face quadrature object
  QUADRATURE* GetVertFaceQuad(int order)
  {
    assert_debug(order>=0 && order<=max_order);
    return &storage[vert_face_quad[order]];
  }

  // Parameters: order - order of the quadrature rule
  // Return value: pointer to the corresponding prism node quadrature object
  QUADRATURE* GetPrismNodeQuad() {return &storage[prism_node_quad];
  }
};

#endif
