//----------------------------------------------------------------------------------------
// recording_stations.H: Definitioins for recording station storage classes
//----------------------------------------------------------------------------------------

#ifndef RECORDING_STATIONS_H
#define RECORDING_STATIONS_H

#include <iostream>
#include <fstream>
#include <vector>
#include <cassert>
#include "general.H"
#include "dof.H"


//----------------------------------------------------------------------------------------
// StationData structure: storage for data on one station
//----------------------------------------------------------------------------------------

struct StationData
{
  real coord[d_space]; // Station coordinates
  int element;         // element id of the element containing the station
};



//----------------------------------------------------------------------------------------
// RecordingStations template class: storage for all recording stations
//----------------------------------------------------------------------------------------

template<int n_var, class T, class LIST>
class RecordingStations
{
private:
  int n_stations;                   // Number of stations
  int output_freq, output_counter;  // Output frequency and counter
  long int output_position;         // Output position in the stations output file
  real start_t, finish_t;           // Starting and end time of stations output
  const char *output_file;          // Name of the stations output file
  std::ofstream *out_stream;        // Output stream
  std::vector<StationData> station; // Station storage array
  VAR var_ind[n_var];               // Idices of monitored variables
  const LIST* list;                 // Pointer to list of elements
  bool verbose;                     // Additional debug info switch

public:
  // Default constructor
  RecordingStations() {out_stream=0; output_position=0; output_counter=1; verbose=false;}

  // Default destructor
  ~RecordingStations() {if (out_stream!=0) {out_stream->close(); delete out_stream;}}

  // Initialize recording stations array
  // Parameters: n_stat - number of stations; freq - frequency of output;
  //             s_time - start time of output; f_time - end time of output;
  //             output - output file name; ind - idices of monitored variables
  void InitializeStations(int n_stat, int freq, real s_time, real f_time, const char* output, VAR ind[])
  {
    n_stations=n_stat;
    if (n_stations==0)
      return;
    output_freq=freq;
    start_t=s_time*86400.;
    finish_t=f_time*86400.;
    output_file=output;
    station.resize(n_stations);
    out_stream = new std::ofstream(output_file, std::ios::in | std::ios::out | std::ios::ate);
    assert(out_stream!=0);
    if (output_position>0 && (*out_stream).tellp()>output_position)
      (*out_stream).seekp(output_position, std::ios::beg);
    for (int i=0; i<n_var; i++)
      var_ind[i]=ind[i];
  }

  // Assign coordinates to a recording station
  // Parameters: station_number - station index; station_coord - station coordinates
  void AssignCoord(int station_number, const real station_coord[])
  {
    assert((station_number>=0) && (station_number<n_stations));
    for (int i=0; i<d_space; station[station_number].coord[i]=station_coord[i], i++);
  }

  // Assign recording stations to elements
  // Parameters: lst - list of elements
  void AssignStations(const LIST& lst)
  {
    list=&lst;
    for (int i=0; i<n_stations; i++)
    {
      real min_dist=1.e25, dist;

      for (std::size_t j=0; j<list->size(); j++)
      {
        dist=(*list)[j].DistanceToBarycenter(station[i].coord);
        if (dist<min_dist)
        {
          min_dist=dist;
          station[i].element=(*list)[j].ID();
        }
        if (min_dist==0.) break;
      }
      const T &el=(*list)[station[i].element];
      if (verbose && min_dist>0.)
      {
        min_dist=el.DistanceToBoundary(station[i].coord);
        std::cout << " Warning: Station " << i+1 
          << " is outside of the computational domain!\n" 
          << " The distance to the boundary of the closest element is: " 
          << min_dist << "\n The output data will be extrapolated!\n";
      }
    }
    verbose=false;
  }

  // Recording station output
  // Parameters: t - time
  void Output(real t)
  {
    if ((n_stations==0) || (t<start_t) || (t>finish_t))
      return;
    if (--output_counter!=0)
      return;
    output_counter=output_freq;

    // Perform outputs
    *out_stream << " " << t << "  ";
    for (int i=0; i<n_stations; i++)
      for (int j=0; j<n_var; j++)
      {
        func_type basis;
        const T& el=(*list)[station[i].element];
        el.Basis_dof(var_ind[j])->basis_func_phys(station[i].coord, el.Mapping(), basis);
        *out_stream << "  " << el.Var_dof(var_ind[j])*basis;
      }
      *out_stream << std::endl;
  }

  // Hot start input
  friend std::ifstream& operator>>(std::ifstream& in, RecordingStations& stations)
  {
    in.read((char*)&stations.output_counter, sizeof stations.output_counter);
    in.read((char*)&stations.output_position, sizeof stations.output_position);
    return in;
  }

  // Hot start output
  friend std::ofstream& operator<<(std::ofstream& out, const RecordingStations& stations)
  {
    unsigned long int output_position;
    if (stations.n_stations>0)
      output_position=stations.out_stream->tellp();
    out.write((char*)&stations.output_counter, sizeof stations.output_counter);
    out.write((char*)&output_position, sizeof output_position);
    return out;
  }
};

#endif
