//----------------------------------------------------------------------------------------
// test_common.cpp: Forcing functions for test problems
//----------------------------------------------------------------------------------------

#include <iostream>
#include <cmath>
#include "pde.H"
#include "test.H"

extern PROBLEM_DESCRIPTION problem_info;

// Forcing function for the test problem
void solution_analytic_forcing(real t, const real coord[d_space], real result[d_var])
{
  static const VAR* ind=VAR_3D_DT_INDEX, *ind_t=VAR_TRANSPORT_INDEX;
  real t_deriv[d_var], var[d_var], grad[d_var][d_space], lapl[d_var][d_space], elev_grad[d_space_2d], 
    diff_coef[d_var][d_space], diff_coef_grad[d_var][d_space]={{0.}}, rhs_diff[d_var]={0.};

  solution_analytic(t, coord, var);
  solution_analytic_dt(t, coord, t_deriv);
  solution_analytic_grad(t, coord, grad);
  solution_analytic_laplacian(t, coord, lapl);
  solution_analytic_elevation_grad(t, coord, elev_grad);
  // Add advection terms
  result[U]=t_deriv[U]+problem_info.g*elev_grad[0]+problem_info.advect_coef*
    (var[U]*(2.*grad[U][0]+grad[V][1]+grad[W][2])+var[V]*grad[U][1]+var[W]*grad[U][2]);
  result[V]=t_deriv[V]+problem_info.g*elev_grad[1]+problem_info.advect_coef*
    (var[V]*(grad[U][0]+2.*grad[V][1]+grad[W][2])+var[U]*grad[V][0]+var[W]*grad[V][2]);
  // Add transport terms
  for (int i=0; i<d_var_transport; i++)
    result[ind_t[i]]=t_deriv[ind_t[i]]+var[ind_t[i]]*(grad[U][0]+grad[V][1]+grad[W][2])+
      var[U]*grad[ind_t[i]][0]+var[V]*grad[ind_t[i]][1]+var[W]*grad[ind_t[i]][2];
  // Add diffusion terms
  diffusion_vert_coef(t, coord, var, grad, diff_coef, diff_coef_grad, rhs_diff);
  for (int i=0; i<problem_info.n_diff_vert; i++)
  {
    VAR ind_diff=problem_info.diff_vert_ind[i];
    result[ind_diff]-=diff_coef[ind_diff][2]*lapl[ind_diff][2]+diff_coef_grad[ind_diff][2]*grad[ind_diff][2];
  }
  for (int i=0; i<problem_info.n_diff_horiz; i++)
  {
    VAR ind_diff=problem_info.diff_horiz_ind[i];
    for (int j=0; j<d_space_2d; j++)
      result[ind_diff]-=problem_info.diff_horiz[ind_diff]*lapl[ind_diff][j];
  }

#ifdef BAROCLINIC_KEY  // Density forcing term
  real surface_elevation, density_forcing_term[d_space_2d];

  solution_analytic_elevation(t, coord, &surface_elevation);
  density_forcing(t, surface_elevation, elev_grad, coord, density_forcing_term);
  result[U]+=density_forcing_term[0];
  result[V]+=density_forcing_term[1];
#endif
#if TURBULENCE_KEY>1  // one equation turbulence models
  result[K]-=problem_info.ramp_value*rhs_diff[K];
#if TURBULENCE_KEY==3
  result[E]-=problem_info.ramp_value*rhs_diff[E];
#endif
#endif
}

// Forcing function for the elevation
void solution_analytic_elevation_forcing(real t, const real x[d_space], real result[])
{
  real coord[d_space]={x[0], x[1]}, var[d_var], elev_dt, elev_grad[d_space_2d];

  solution_analytic_elevation(t, coord, coord+2);
  solution_analytic(t, coord, var);
  solution_analytic_elevation_grad(t, coord, elev_grad);
  solution_analytic_elevation_dt(t, coord, &elev_dt);
  result[0]=(elev_dt+var[U]*elev_grad[0]+var[V]*elev_grad[1]-var[W]);
}

// Initial condition
void initial_condition(real t, const real coord[d_space], real result[d_var])
{
  solution_analytic(t, coord, result);
}

// Initial condition for elevation
void initial_elevation(real t,  const real coord[d_space_2d], real result[])
{
  solution_analytic_elevation(t, coord, result);
}
