//----------------------------------------------------------------------------------------
// Analytical solution for test problems
//----------------------------------------------------------------------------------------

#include <cmath>
#include "pde.H"
#include "test.H"

extern PROBLEM_DESCRIPTION problem_info;

const long_real freq=0.01, elev_coef=0.01, amp=0.1, t_coef=1.;

// True solution for the test problem
void solution_analytic(real t, const real coord[d_space], real func[d_var])
{
  long_real s0=std::sin(freq*(coord[0]+t_coef*t)), s1=std::sin(freq*(coord[1]+t_coef*t)), 
    c0=std::cos(freq*(coord[0]+t_coef*t)), c1=std::cos(freq*(coord[1]+t_coef*t)), 
    zb=0.005*(coord[0]+coord[1])-2., coef=-amp*(coord[2]-zb);

  func[U]=coef*s0;
  func[V]=coef*s1;
  func[W]=coef*(0.005*(s0+s1)-0.5*freq*(coord[2]-zb)*(c0+c1));
#ifdef BAROCLINIC_KEY
  func[T]=20.;
  func[S]=35.;
  func[RO]=density(problem_info, coord, func);
#endif
}

// Gradient of the true solution
void solution_analytic_grad(real t, const real coord[d_space], real grad[d_var][d_space])
{
  long_real s0=std::sin(freq*(coord[0]+t_coef*t)), s1=std::sin(freq*(coord[1]+t_coef*t)), 
    c0=std::cos(freq*(coord[0]+t_coef*t)), c1=std::cos(freq*(coord[1]+t_coef*t)), 
    zb=0.005*(coord[0]+coord[1])-2., coef=-amp*freq*(coord[2]-zb);

  grad[U][0]=coef*c0+amp*0.005*s0;
  grad[U][1]=amp*0.005*s0;
  grad[U][2]=-amp*s0;
  grad[V][0]=amp*0.005*s1;
  grad[V][1]=coef*c1+amp*0.005*s1;
  grad[V][2]=-amp*s1;
  grad[W][0]=coef*(0.005*c0+0.5*freq*(coord[2]-zb)*s0);
  grad[W][1]=coef*(0.005*c1+0.5*freq*(coord[2]-zb)*s1);
  grad[W][2]=-amp*0.005*(s0+s1)-coef*(c0+c1);
#ifdef BAROCLINIC_KEY
  grad[T][0]=0.;
  grad[T][1]=0.;
  grad[T][2]=0.;
  grad[S][0]=0.;
  grad[S][1]=0.;
  grad[S][2]=0.;
#endif
}

// Laplacian of the true solution
void solution_analytic_laplacian(real t, const real coord[d_space], real lapl[d_var][d_space])
{
  long_real s0=std::sin(freq*(coord[0]+t_coef*t)), s1=std::sin(freq*(coord[1]+t_coef*t)), 
    c0=std::cos(freq*(coord[0]+t_coef*t)), c1=std::cos(freq*(coord[1]+t_coef*t)), 
    zb=0.005*(coord[0]+coord[1])-2., coef=amp*freq;

  lapl[U][0]=coef*(freq*(coord[2]-zb)*s0+0.01*c0);
  lapl[U][1]=0.;
  lapl[U][2]=0.;
  lapl[V][0]=0.;
  lapl[V][1]=coef*(freq*(coord[2]-zb)*s1+0.01*c1);
  lapl[V][2]=0.;
#ifdef BAROCLINIC_KEY
  lapl[T][0]=0.;
  lapl[T][1]=0.;
  lapl[T][2]=0.;
  lapl[S][0]=0.;
  lapl[S][1]=0.;
  lapl[S][2]=0.;
#endif
}

// Time derivative of the true solution for the test problem
void solution_analytic_dt(real t, const real coord[d_space], real t_deriv[d_var])
{
  long_real zb=0.005*(coord[0]+coord[1])-2., coef=-amp*freq*t_coef*(coord[2]-zb);

  t_deriv[U]=coef*std::cos(freq*(coord[0]+t_coef*t));
  t_deriv[V]=coef*std::cos(freq*(coord[1]+t_coef*t));
#ifdef BAROCLINIC_KEY
  t_deriv[T]=0.;
  t_deriv[S]=0.;
#endif
}

// Analytic solution for the elevation
void solution_analytic_elevation(real t, const real coord[d_space_2d], real result[])
{
  result[0]=elev_coef*(std::sin(freq*(coord[0]+t_coef*t))+std::sin(freq*(coord[1]+t_coef*t)));
}

// Analytic solution for the elevation gradient
void solution_analytic_elevation_grad(real t, const real coord[d_space_2d], real elev_grad[d_space_2d])
{
  elev_grad[0]=elev_coef*freq*std::cos(freq*(coord[0]+t_coef*t));
  elev_grad[1]=elev_coef*freq*std::cos(freq*(coord[1]+t_coef*t));
}

// Time derivative of the true solution for the elevation
void solution_analytic_elevation_dt(real t, const real coord[d_space], real result[])
{
  result[0]=t_coef*elev_coef*
    freq*(std::cos(freq*(coord[0]+t_coef*t))+std::cos(freq*(coord[1]+t_coef*t)));
}

#ifdef BAROCLINIC_KEY
long_real density(const real coord[d_space], const real var[d_var])
{
  return 0.1*coord[0];
}

// Forcing function accounting for density terms
void density_forcing(real t, real xi, const real xi_grad[d_space_2d], const real coord[d_space], real result[])
{
  real coord_top[d_space]={coord[0], coord[1], xi}, var[d_var], ro=density(coord_top, var);

  result[0]=0.1*(xi-coord[2])+problem_info.g*ro*xi_grad[0];
  result[1]=problem_info.g*ro*xi_grad[1];
}
#endif
